# -*- coding:utf-8 -*-
# 文件:  thread_pool.py
# 日期:  2022/8/19 9:50
"""

"""
import datetime
import threading
from concurrent import futures
from .logs import logger


__version__ = '1.4'
__all__ = ['JdThreadPool']


"""
Future 提供了如下方法：
cancel()：取消该 Future 代表的线程任务。如果该任务正在执行，不可取消，则该方法返回 False；否则，程序会取消该任务，并返回 True。
cancelled()：返回 Future 代表的线程任务是否被成功取消。
running()：如果该 Future 代表的线程任务正在执行、不可被取消，该方法返回 True。
done()：如果该 Future 代表的线程任务被成功取消或执行完成，则该方法返回 True。
result(timeout=None)：获取该 Future 代表的线程任务最后返回的结果。如果 Future 代表的线程任务还未完成，该方法将会阻塞当前线程，其中 timeout 参数指定最多阻塞多少秒。
exception(timeout=None)：获取该 Future 代表的线程任务所引发的异常。如果该任务成功完成，没有异常，则该方法返回 None。
add_done_callback(fn)：为该 Future 代表的线程任务注册一个“回调函数”，当该任务成功完成时，程序会自动触发该 fn 函数。
"""

# 0-未启动， 1-running, 2-canceled, 3-done
TASK_NOT_START = 0
TASK_RUNNING = 1
TASK_CANCELED = 2
TASK_DONE = 3


class JdTaskStatus(object):
    # 任务状态
    def __init__(self, name, future):
        self.name = name    # 任务名称
        self.start_time = datetime.datetime.today()  # 任务启动时间
        self.finish_time = None  # 任务完成时间
        self.status = 0     # 0-未启动， 1-running, 2-canceled, 3-done
        self.future = future    # 任务


class JdThreadPool(object):
    """
    线程池
    """

    def __init__(self, max_workers=5):
        """
        创建线程池
        """
        self.pool = futures.ThreadPoolExecutor(max_workers=max_workers)
        self._thread_lock = threading.Lock()
        self._future = {}  # future -> name
        self._task = {}   # name -> JdTaskStatus

    def add_job(self, callback, para=None, name=None) -> futures.Future:
        """
        添加任务
        callback:   任务入口函数
        para:       上述入口函数的参数， None时不传递
        name:       任务名称， 用于防止任务重复执行
        """
        thread_name = threading.current_thread().name
        logger.debug(f"[ {thread_name} ], {para}")
        
        if name and name in self._task:  # 任务已经添加过，且是运行状态时， 不再添加重复任务
            t = self._task[name]
            if isinstance(t, JdTaskStatus) and isinstance(t.future, futures.Future):
                if t.future.running():
                    return t.future
        
        if para is not None:
            future = self.pool.submit(callback, para)
        else:
            future = self.pool.submit(callback)
        self._thread_lock.acquire()
        self._future[future] = name or ''
        if name:
            self._task[name] = JdTaskStatus(name, future)
        self._thread_lock.release()
        future.add_done_callback(self._callback_finished)
        return future

    def _callback_finished(self, future: futures.Future):
        """
        线程执行完毕后的 回调函数
        """
        thread_name = threading.current_thread().name
        logger.debug(f"[ {thread_name} ]执行完成： {future.result()}")
        
        self._thread_lock.acquire()
        name = self._future.pop(future, None)
        if name in self._task:
            t = self._task[name]
            t.status = TASK_DONE    # 任务已经完成
        self._thread_lock.release()

    @classmethod
    def cancel(cls, future: futures.Future):
        """
        取消该 Future 代表的线程任务。如果该任务正在执行，不可取消，则该方法返回 False；
        否则，程序会取消该任务，并返回 True。
        """
        return future.cancel()

    @classmethod
    def cancelled(cls, future: futures.Future):
        """
        返回 Future 代表的线程任务是否被成功取消。
        """
        return future.cancelled()

    @classmethod
    def running(cls, future: futures.Future):
        """
        如果该 Future 代表的线程任务正在执行、不可被取消，该方法返回 True。
        """
        return future.running()

    @classmethod
    def done(cls, future: futures.Future):
        """
        如果该 Future 代表的线程任务被成功取消或执行完成，则该方法返回 True。
        """
        return future.done()
    
    def is_running(self, name):
        # 任务 name 是否处于运行状态  True-运行中 False-非运行中，已经结束 或 取消状态
        
        t = self._task.get(name)
        if t and isinstance(t, JdTaskStatus) and isinstance(t.future, futures.Future):
            return t.future.running()
        return False



# -*- coding: utf-8 -*-
from enum import Enum, unique
import json
import requests
from urllib3.exceptions import MaxRetryError, NewConnectionError
import base64
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

__version__ = '2.9.8'
__all__ = ['RequestType', 'Core', 'aes_decrypt', 'jd_concat']


@unique
class RequestType(Enum):
    GET = 'GET'
    POST = 'POST'
    PATCH = 'PATCH'
    PUT = 'PUT'
    DELETE = 'DELETE'


_timeout = (3.05, 27)   # 超时时间, 连接超时 和 读取超时
# _timeout = None


class Core:
    def __init__(self,
                 api_key=None,      # 秘钥 key
                 logger=None,       # 日志
                 proxy=None,        # 代理
                 gate_way='',       # 网关， 网关 + path 构成请求接口地址
                 ):
        self._proxy = proxy
        self._api_key = api_key
        self._gate_way = gate_way
        self._logger = logger

    def request(self, path, method=RequestType.GET, data=None, files=None, headers=None,
                show_url=True, show_response=True, force_data=False,
                show_request=True, show_rsp=True):
        """
        请求 POST 、GET
        :param path:            接口 路径， 网关 + path 构成请求接口地址
        :param method:          POST、GET(default method)
        :param data:            请求数据
        :param files:           请求文件
        :param headers:         请求头
        :param show_request:    是否显示请求信息，包括 url， POST/GET类型，头部，请求参数。 True-显示， False-不显示
        :param show_url:        是否在日志显示 url。 优先级小于 show_request
        :param show_rsp:        是否在日志显示接口返回所有内容（状态码，header，text）。 True-显示， False-不显示
        :param show_response:   是否在日志显示接口返回内容。优先级小于 show_rsp
        :param force_data:      按 字节方式传递 data
        :return:    (status_code, message)
        """
        if headers is None:
            headers = {}
        if files:
            headers.update({'Content-Type': 'multipart/form-data'})
        else:
            headers.update({'Content-Type': 'application/json'})
        headers.update({'Accept': 'application/json'})
        headers.update({'User-Agent': 'jd-wechatpay v3 python sdk'})
        netloc = path.find("://", 4, 8)  # http:// or https://.  '://' start begin 4 or 5
        if netloc > 0:  # path 为完成 url，含有域名
            url = path
        else:  # path 为路径
            url = jd_concat(self._gate_way, path)   # 网关 + 路径 构成 发送 url
        if self._logger and show_request:  # 显示请求信息
            if show_url:
                self._logger.debug('Request url: %s' % url)
            self._logger.debug('Request type: %s' % method.value)
            self._logger.debug('Request headers: %s' % headers)
            self._logger.debug('Request params: %s' % data)
        try:
            if method == RequestType.GET:
                response = requests.get(url=url, params=data, headers=headers, proxies=self._proxy, timeout=_timeout)
            elif method == RequestType.POST:
                para_json = None if force_data or files else data
                para_byte = data if force_data or files else None
                if force_data and isinstance(para_byte, (str, dict)):   # 字符串、字典 转 byte
                    para_byte = json.dumps(para_byte, ensure_ascii=False).encode("utf-8")
                response = requests.post(url=url, json=para_json, data=para_byte, headers=headers, files=files,
                                         proxies=self._proxy, timeout=_timeout)
            elif method == RequestType.PATCH:
                response = requests.patch(url=url, json=data, headers=headers, proxies=self._proxy, timeout=_timeout)
            elif method == RequestType.PUT:
                response = requests.put(url=url, json=data, headers=headers, proxies=self._proxy, timeout=_timeout)
            elif method == RequestType.DELETE:
                response = requests.delete(url=url, headers=headers, proxies=self._proxy, timeout=_timeout)
            else:
                raise Exception('sdk does no support this request type.')
        except (TimeoutError, requests.exceptions.ConnectionError, MaxRetryError, NewConnectionError):
            return 500, {'code': 500, 'msg': 'Timeout Error'}
        if self._logger and show_request and show_url:  # 真实 url
            self._logger.debug('--- Request [real] url: %s' % response.url)
        if self._logger and show_rsp:
            self._logger.debug('Response status code: %s' % response.status_code)
            self._logger.debug('Response headers: %s' % response.headers)
            if show_response:
                self._logger.debug('Response content: %s' % response.text)
        if response.status_code == 404:
            return response.status_code, {'code': response.status_code, 'msg': response.text}
        elif response.status_code in [405]:
            return response.status_code, {'code': response.status_code, 'msg': 'Method Not Allowed'}
        content_type = response.headers.get('Content-Type')
        if content_type in ['image/jpeg', 'image/png', 'image/gif', 'video/mp4', 'audio/mp3', 'video/mpeg']:
            message = response.content
        else:
            message = response.text if content_type and 'application/json' in content_type else response.content
            try:
                if isinstance(message, str):
                    message = json.loads(message)
                elif isinstance(message, bytes):
                    message = json.loads(message.decode("utf-8"))
            except json.decoder.JSONDecodeError:
                message = {}
            except UnicodeDecodeError:
                pass
        return response.status_code, message

    def decrypt_callback(self, headers, body):
        """
        通知的数据会加密 post 到对接方接口中，由于涉及到回调加密和解密，接入方必须先通过微校商务侧申请
        通知秘钥 key，申请好通知秘钥 key 后才能解密回调通知
        :param headers:
        :param body:
        :return:
        """
        if isinstance(body, bytes):
            body = body.decode('UTF-8')
        if self._logger:
            self._logger.debug('Callback Header: %s' % headers)
            self._logger.debug('Callback Body: %s' % body)
        data = json.loads(body)
        resource_type = data.get('resource_type')
        if resource_type != 'encrypt-resource':
            return None
        resource = data.get('resource')
        if not resource:
            return None
        algorithm = resource.get('algorithm')
        if algorithm != 'AEAD_AES_256_GCM':
            raise Exception('sdk does not support this algorithm')
        nonce = resource.get('nonce')
        ciphertext = resource.get('ciphertext')
        associated_data = resource.get('associated_data')
        if not (nonce and ciphertext):
            return None
        if not associated_data:
            associated_data = ''
        result = aes_decrypt(
            nonce=nonce,
            ciphertext=ciphertext,
            associated_data=associated_data,
            notify_key=self._api_key)
        if self._logger:
            self._logger.debug('Callback resource: %s' % result)
        return result


def jd_concat(url: str, path: str):
    """
    url 合并， 去掉重复的 / 字符
    :param url:     url 网关，例如： 'https://domain.com
    :param path:    请求路径，例如： '/some_path/books_get'
    :return:    合并后的路径  url + path，去掉 多余 字符('/')
    """
    if url and path:
        n = 1 if url[-1] == '/' else 0
        if path[0] == '/':
            n += 1

        if 2 == n:
            return url + path[1:]
        elif 0 == n:
            return f"{url}/{path}"

    return (url or '') + (path or '')


def aes_decrypt(nonce, ciphertext, associated_data, notify_key):
    """
    微卡 订单流水实时推送第三方平台， 数据解密

    下面详细描述对通知数据进行解密的流程：
    1. 通知秘钥 key，记为 notify_key；
    2. 针对 resource.algorithm 中描述的算法（目前为 AEAD_AES_256_GCM），取得对应的参数 nonce 和 associated_data；
    3. 使用 notify_key、nonce 和 associated_data，对数据密文 resource.ciphertext 进行解密，得到 JSON 形式的资源
    对象（资源对象就是具体的数据内容）；
    注： AEAD_AES_256_GCM 算法的接口细节，请参考 rfc5116。微校支付使用的通知密钥 key 长度为 32 个字节，随机串 nonce
    长度 32 个字节，associated_data 长度小于 16 个字节并可能为空。
    :param nonce:               加密使用的随机串
    :param ciphertext:          Base64编码后的数据密文
    :param associated_data:     附加数据
    :param notify_key:          秘钥
    :return:
    """
    key_bytes = notify_key.encode('UTF-8')
    nonce_bytes = nonce.encode('UTF-8')
    associated_data_bytes = associated_data.encode('UTF-8') if associated_data else None
    data = base64.b64decode(ciphertext)
    aes_gcm = AESGCM(key=key_bytes)
    try:
        result = aes_gcm.decrypt(nonce=nonce_bytes, data=data, associated_data=associated_data_bytes).decode('UTF-8')
    except InvalidTag:
        result = None
    return result



# -*- coding: utf8 -*-
# 文件:  tools.py
# 日期:  2021/10/7 14:02
"""
DATE        AUTHOR  DESCRIPTION
----------  ------  -----------------------------------------------------------------
2022-07-02  WXG     JdTools.str_to_datetime 支持格式 RFC 3339 , 版本改为 2.1
"""
import time
import datetime
import calendar
import os
import operator
import re
import xlrd
from PIL import Image
import qrcode
from decimal import Decimal, ROUND_HALF_UP, InvalidOperation, DecimalException
from .crypto import get_nonce


__version__ = '2.5.1'

__all__ = [
    'dc_gen_code', 'dc_records_changed', 'start_end_time', 'start_end_time_ex',
    'JdCalendar', 'JdTools', 'JdDate', 'JdImage', 'JdList', 'JdPath', 'JdFmt',
    'DT_YM', 'DT_SHORT', 'DT_LONG_LESS', 'DT_LONG', 'DT_FULL',
    'DT_MD', 'DT_FULL_EX', 'DT_FULL_EX2',
    'name_list', 'sum_list', 'sum_ticket', 'Singleton', 'humanize_date_delta',
    'singleton'
]


def dc_gen_code(school_no, tag, date=None):
    """
    生成单据编号，包括学号，班级编号，收费单号等。
    :param school_no:       分校编号，例如：0001,0002
    :param tag:             单据标识，例如学号为 XH， 收费单为 SFD
    :param date:            日期，单据上附加日期信息
    :return:        生成的编号, 例如 0001-XH-181008-       [完整单据号：0001-XH-181008-001]
    """
    if date is None:
        date = datetime.datetime.today()
    data_str = datetime.datetime.strftime(date, '%y%m%d-')
    code_str = '%04d-%s-%s' % (int(school_no), tag, data_str)
    return code_str


def dc_records_changed(old: list, new: list, field='id'):
    """修改记录（全量）， 判断在原纪录基础上的删、改、增情况。
       var aa = [{id: 1, name:'Tom'},{id:2, name:'Peter'}]; 原始记录
       var bb = [{name:'Alice'}, {id:2, name: 'PP'}];       最终记录。
       var chg = dc_records_changed(aa, bb, 'id')
       则需要增加bb中的第一条，修改为bb中第二条，删除aa中第一个条。
       返回值为 {add:[0], del:[0], upd:[1]}
    :param old:     原始记录 [{}, {}]
    :param new:     修改后的记录  [ {}, {}]
    :param field:   比较字段，用于判断增、改、删
    :return:    {'add': list, 'del': list, 'upd': list}
        add -- new 中 下标
        upd -- new 中 下标
        del -- old 中 下标
    """
    add_idx = []
    del_idx = []
    upd_idx = []
    upd_idx_old = []    # old update index
    ori = set()
    cur = set()

    for i in old:
        if field in i:
            ori.add(i[field])
    for i in range(len(new)):
        if field in new[i]:
            cur.add(new[i][field])
        else:
            add_idx.append(i)
    del_key = ori.difference(cur)
    upd_key = ori.intersection(cur)
    add_key = cur.difference(ori)

    # 找出 del_key, upd_key, add_key  对应的下标
    for i in range(len(old)):
        if field in old[i] and old[i][field] in del_key:
            del_idx.append(i)
    for i in range(len(new)):
        if field in new[i] and new[i][field] in upd_key:
            upd_idx.append(i)
            for j in range((len(old))):
                if field in old[j] and old[j][field] == new[i][field]:
                    upd_idx_old.append(j)   # 记录 old 中 和 new 对应 key 的索引
    for i in range(len(new)):
        if field in new[i] and new[i][field] in add_key:
            add_idx.append(i)

    return {'add': add_idx, 'del': del_idx, 'upd': upd_idx, 'upd_old': upd_idx_old}


def start_end_time(tm, val=None):
    """
    返回 日期 tm，对应的开始和结束日期
    :param tm:  日期名：today, yesterday, this-week, last-week, this-month, last-month,
                    year-month, time_range
    :param val:  日期字符串， 当为 year-month 时： 年-月。 time_range时：年-月-日 - 年-月-日
    :return: 开始日期 和 结束日期
    """
    one_day = datetime.timedelta(hours=23, minutes=59, seconds=59, microseconds=999999)
    if tm == 'year-month':      # 按照 年 月 查询
        ym = val.split('-')     # 年-月： 2019-09
        year = int(ym[0])
        month = int(ym[1])
        d1 = datetime.datetime(year=year, month=month, day=1)
        month_range = calendar.monthrange(year, month)
        d2 = d1 + datetime.timedelta(days=month_range[1]-1) + one_day
    elif tm == 'today':     # 今天
        td = datetime.date.today()
        d1 = datetime.datetime.strptime(str(td), '%Y-%m-%d')
        d2 = d1 + one_day
    elif tm == 'yesterday':
        td = datetime.date.today()
        d1 = datetime.datetime.strptime(str(td), '%Y-%m-%d') - datetime.timedelta(days=1)
        d2 = d1 + one_day
    elif tm == 'this-week':
        cal = JdCalendar()
        # cal.set_monday_is_first(False)
        d1, d2 = cal.this_week()
    elif tm == 'last-week':
        cal = JdCalendar()
        # cal.set_monday_is_first(False)
        d1, d2 = cal.last_week()
    elif tm == 'this-month':
        cal = JdCalendar()
        d1, d2 = cal.this_month()
    elif tm == 'last-month':
        cal = JdCalendar()
        d1, d2 = cal.last_month()
    elif tm == 'time_range':
        times = val.split(' - ')
        if len(val) <= 17:  # val 格式 2020-01 - 2020-12
            ym = times[0].split('-')
            d1 = datetime.datetime(year=int(ym[0]), month=int(ym[1]), day=1)
            ym = times[1].split('-')
            d = datetime.datetime(year=int(ym[0]), month=int(ym[1]), day=1)
            month_range = calendar.monthrange(int(ym[0]), int(ym[1]))
            d2 = d + datetime.timedelta(days=month_range[1] - 1) + one_day
        else:   # val 格式 2019-11-01 - 2019-11-17
            d1 = datetime.datetime.strptime(JdTools.trim(times[0]), '%Y-%m-%d')
            d2 = datetime.datetime.strptime(JdTools.trim(times[1]), '%Y-%m-%d') + one_day
    else:
        d1 = datetime.datetime.strptime(tm, '%Y-%m-%d')
        d2 = d1 + one_day
    return d1, d2


def start_end_time_ex(tm, val=None):
    """
    返回 日期 tm，对应的开始和结束日期，带时间段描述
    :param tm:  日期，见 start_end_time
    :param val: 日期字符串， 当为 year-month 时： 年-月。 time_range时：年-月-日 - 年-月-日
    :return:  开始日期, 结束日期, 开始到结束日期的文字描述
    """
    d1, d2 = start_end_time(tm, val)
    if d1.date() == d2.date():  # 今天或者昨天的时间段
        date_from_to = '%s' % (JdTools.date_ymd(d1))
    else:
        date_from_to = '从 %s 至 %s' % (JdTools.date_ymd(d1), JdTools.date_ymd(d2))
    return d1, d2, date_from_to


class JdCalendar(object):
    def __init__(self, date_time=None, monday_is_first=True):
        self.date_time = date_time if date_time else datetime.datetime.today()
        self.date = datetime.datetime.strptime(str(self.date_time.date()), '%Y-%m-%d')
        self.iso_weekday = self.date.isoweekday()
        self.weekday = self.date.weekday()
        self.one_day = datetime.timedelta(hours=23, minutes=59, seconds=59, microseconds=999999)
        self.monday_is_first = monday_is_first

    def last_week(self):
        date_to = self.date - datetime.timedelta(days=self.iso_weekday)
        date_from = date_to - datetime.timedelta(days=6)
        date_to += self.one_day
        if self.monday_is_first is False:
            date_from -= datetime.timedelta(days=1)
            date_to -= datetime.timedelta(days=1)
        return date_from, date_to

    def this_week(self):
        date_from = self.date - datetime.timedelta(days=self.iso_weekday-1)
        date_to = date_from + self.one_day + datetime.timedelta(days=6)
        if self.monday_is_first is False:
            date_from -= datetime.timedelta(days=1)
            date_to -= datetime.timedelta(days=1)
        return date_from, date_to

    def set_monday_is_first(self, is_first=True):
        self.monday_is_first = is_first

    def this_month(self):
        year = self.date.year
        month = self.date.month
        date_from = datetime.datetime(year, month, 1, 0, 0, 0)
        month_range = calendar.monthrange(year, month)
        date_to = date_from + datetime.timedelta(days=month_range[1] - 1) + self.one_day
        return date_from, date_to

    def last_month(self):
        month = 12 if self.date.month == 1 else self.date.month - 1
        year = self.date.year if month != 12 else self.date.year - 1
        date_from = datetime.datetime(year, month, 1, 0, 0, 0)
        month_range = calendar.monthrange(year, month)
        date_to = date_from + datetime.timedelta(days=month_range[1] - 1) + self.one_day
        return date_from, date_to


# 日期格式
DT_YM = 5  # 年-月
DT_SHORT = 1  # 年-月-日 y4-mm-dd
DT_LONG_LESS = 3  # 年-月-日 时:分 y4-mm-dd HH:MM
DT_LONG = 2  # 年-月-日 时:分:秒 y4-mm-dd HH:MM:SS
DT_FULL = 4  # 年-月-日 时:分:秒.毫秒 y4-mm-dd HH:MM:SS.ms
DT_MD = 6       # 月-日 mm-dd
DT_FULL_EX = 12  # 年月日-时分秒-毫秒
DT_FULL_EX2 = 13  # 年月日_时分秒_毫秒


class JdTools(object):
    """工具类，实现常用函数"""
    @staticmethod
    def getdate(no_ms=True, nonce=True):
        """获取当前日期字符串 返回 年月日_时分秒 例如：20180921_084800_ab67DE
        :param no_ms:       文件名不带毫秒
        :param nonce:       文件名是否带随机字符
        """
        return JdTools.get_date(no_ms, nonce)

    @staticmethod
    def get_date(no_ms=True, nonce=True):
        """获取当前日期字符串 返回 年月日_时分秒_6位随机字符 例如：20180921_084800_ab67DE
        :param no_ms:       文件名不带毫秒
        :param nonce:       文件名是否带随机字符
        """
        tm = datetime.datetime.today()
        fmt = '%Y%m%d_%H%M%S' if no_ms else '%Y%m%d_%H%M%S_%f'
        data_str = datetime.datetime.strftime(tm, fmt)
        if nonce:
            return f'{data_str}_{get_nonce(6)}'
        return f'{data_str}'

    @staticmethod
    def trim(s):
        """过滤字符串首尾空格，例如   '  Hello World  ' ->  'Hello World' """

        """
        s = s.strip()
        """
        space = [' ', '\t', '\n', '\r']
        if not isinstance(s, str) and not isinstance(s, bytes):
            return s
        while s and s[0] in space:
            s = s[1:]
        while s and s[-1] in space:
            s = s[:-1]

        return s

    @staticmethod
    def split(s, sep, num=-1):
        """分割字符串，并过滤首尾空格 """
        if not isinstance(s, str) and not isinstance(s, bytes):
            return s
        arr = s.split(sep, num)
        arr = [JdTools.trim(i) for i in arr]
        return arr

    @staticmethod
    def str2arr(s):  # 字符串（用英文逗号分隔） 转 列表。 例如：'1,2,3' -> [1, 2, 3]， 字符串非法，不报错
        if not isinstance(s, str) and not isinstance(s, bytes):
            return []
        arr = JdTools.split(s, ',')
        ret = []
        for i in arr:
            passed, v = JdTools.int(i)
            if passed:
                ret.append(v)
        return ret

    @staticmethod
    def get_year_month(str_date):
        """从字符串获取年和月份， str_date 字符串格式 '年-月-日 时:分:秒' """
        arr = str_date.split('-')
        if len(arr) < 2:
            raise Exception('str_date 格式错误，期望【年-月-日 时:分:秒】')
        year = int(arr[0])
        month = int(arr[1])
        return year, month

    @staticmethod
    def float_equal(f1, f2, precision=1e-5):
        """判断浮点数 f1 和 f2 是否相等。precision 为比较精度，默认为 1e-5 (10的-5次方) """
        if f1 is None and f2 is None:   # 都为None 返回 True
            return True
        if f1 is None or f2 is None:    # 只有一个是 None，返回 False
            return False
        return abs(float(f1)-float(f2)) <= precision

    @staticmethod
    def get_year_month_day(str_date):
        """从字符串获取年和月份和日， str_date 字符串格式 '年-月-日 时:分:秒' """
        arr = str_date.split('-')
        if len(arr) < 3:
            raise Exception('str_date 格式错误，期望【年-月-日 时:分:秒】')
        return arr[0], arr[1], arr[2]

    @staticmethod
    def list_eq(x, y):
        return operator.eq(x, y)    # Python 3  True or False

    @staticmethod
    def get_filename(name, excel2010=False, no_ext=False, no_ms=True, nonce=True):
        """
        返回 唯一 文件名，name 后面增加 年月日_时分秒_毫秒 后缀
        :param name:        文件名前缀
        :param excel2010:   excel文件后缀格式， True-4位后缀文件扩展名， False - xls
        :param no_ext:      不带文件名后缀
        :param no_ms:       文件名不带毫秒
        :param nonce:       文件名是否带随机字符
        :return:    文件名 name + 日期 + 扩展名
        """
        time_str = JdTools.get_date(no_ms=no_ms, nonce=nonce)
        ext = '' if no_ext else ('.xls' + ('x' if excel2010 else ''))
        return name + time_str + ext

    @staticmethod
    def str(val):
        """ int to str， val 为 [0, 9]时 返回 '01', '02', ..., '09' """
        if not isinstance(val, int):
            raise Exception('JdTools.str expect int.')
        v = '0' + str(val) if 0 <= val <= 9 else str(val)
        return v

    @staticmethod
    def get_end_time(begin, minutes):
        """
        上课开始时间，上课时长，转 开始结束时间

        :param begin:       上课开始时间，例如 09:00
        :param minutes:     上课时长，单位分钟，例如 90
        :return:    开始时间-结束时间，例如：09:00-10:30
        """
        if isinstance(minutes, float):
            minutes = int(minutes)

        arr = begin.split('-')
        b = arr[0].split(':')

        h = int(b[0]) + minutes // 60
        m = int(b[1]) + minutes % 60
        if m >= 60:
            h += 1
            m -= 60
        begin_end_time = arr[0] + "-" + JdTools.str(h) + ":" + JdTools.str(m)
        return begin_end_time

    @staticmethod
    def is_float(val):
        """判断 val是否可以转为float。可以转换，返回True，不能转换，返回False"""
        try:
            float(val)
            return True
        except (ValueError, TypeError):
            return False

    @staticmethod
    def float(val):
        """将val转为float。转换成功，返回 True 和 转换后的值；转换失败，返回 False, 0"""
        try:
            v = float(val)
            return True, v
        except (ValueError, TypeError):
            return False, 0

    @staticmethod
    def int(val, default=0):
        """将val转为 int。转换成功，返回 True 和 转换后的值；转换失败，返回 False, 0"""
        try:
            v = int(val)
            return True, v
        except (ValueError, TypeError):
            return False, default

    @staticmethod
    def get_user_code(code):
        """格式化用户（操作员）编号"""
        return 'USER-%03d' % int(code)

    @staticmethod
    def get_float(v, num=2, force=False):
        """ float to str,  保留2位小数。当 force 不填/'假', 小数点后全是0（是整数），则转换为整数（不带小数点）"""
        # 123.3456 --> '123.35';    180 -> '180'

        if v is None:  # 兼容性处理
            v = 0
            if not force:
                return v
        elif isinstance(v, str):
            _, v = JdTools.float(v)
        fmt_dot = '{:.%sf}' % num       # '{:.2f}'.format(v)  === '%.2f' % v
        val = fmt_dot.format(v)

        if force:
            return val
        '''
        c = val[-1]
        while c == '0':  # 消除 小数点 末尾 的 0
            val = val[: -1]
            c = val[-1]
        if val[-1] == '.':  # 对于整数，删除 小数点
            val = val[: -1]
        '''
        arr = val.rsplit('.')
        if arr[1] == '0'*num:
            val = arr[0]
        return val

    @staticmethod
    def get_float2(v):
        """返回字符串，带小数点后2位"""
        return JdTools.get_float(v, 2, True)

    @staticmethod
    def money_fmt(v):
        return float(JdTools.get_float(v, 2))

    @staticmethod
    def format_salary(fee, count_method, has_auth=True, num=None, stu_fee=None):
        """
        教师工资算法转字符串，转换结果示例：  100元/人 or 100元/节
        :param fee:     工资算法收费标准，单位 元
        :param count_method:    工资计费方式， 1 每节课固定 2 按学生人头  3-按创收比率计算  4 课次+人头 5-单次阶梯, 6-月带课数阶梯
        :param has_auth:    是否有权限，无权限时，屏蔽 fee 的值
        :param num:    当 count_method=4时，有效。超过人数
        :param stu_fee:    当 count_method=4时，有效。人头费
        :return:  fee对应的数值 元/人 or 元/节。 若 fee, count_method 其中一个为None，则返回 None
        """
        if fee is None or count_method is None:
            return None
        f = JdTools.get_float(fee)
        std = f if has_auth else '*' * len(f)
        if count_method == 1:
            text = '%s元/节' % std
        elif count_method == 2:
            text = '%s元/人' % std
        elif count_method == 3:
            text = '创收%s%%' % std
        elif count_method == 4:
            head = JdTools.get_float(stu_fee)
            head_std = head if has_auth else '*' * len(head)
            text = '%s元/节+%s元/人(>%s人)' % (std, head_std, num)
        elif count_method == 5:
            text = '单次阶梯算法'
        elif count_method == 6:
            text = '月带课数阶梯算法'
        else:
            text = None
        return text

    @staticmethod
    def str_to_datetime(s_dt, add_hms=False):
        """
        字符串 转 日期， datetime.datetime 格式

        RFC 3339格式：“{年}-{月}-{日}T{时}:{分}:{秒}.{毫秒}{时区}”；
        其中的年要用零补齐为4位，月日时分秒则补齐为2位。毫秒部分是可选的。
        最后一部分是时区，前面例子中的 Z 其实是零时区 Zulu 的缩写，它也可能是 +08:00 或 -08:00 等；
        2017-12-08T00:00:00.00Z
        2017-12-08T08:00:00.00+08:00
        都代表所在时区的本地时间。
        ISO8610与RFC3339有各自独特的表示法，也有重合部分。
        :param s_dt:    日期字符串
        :param add_hms: 转换后是否添加时分秒
        :return:
        """
        if isinstance(s_dt, (datetime.datetime, datetime.date)):
            return s_dt

        now = datetime.datetime.today()
        dt = datetime.timedelta(hours=now.hour, minutes=now.minute, seconds=now.second)

        if s_dt == '' or s_dt is None:
            return None
        length = len(s_dt)
        has_dot = True if '.' in s_dt else False

        fmt = '%Y-%m-%d'
        if length <= 7:
            fmt = '%Y-%m'
        elif length <= 10:
            if has_dot:
                fmt = '%Y.%m.%d'
            elif '/' in s_dt:
                fmt = '%Y/%m/%d'
            else:
                fmt = '%Y-%m-%d'
        elif length <= 16:
            fmt = '%Y-%m-%d %H:%M'
        elif length <= 19:
            fmt = '%Y-%m-%d %H:%M:%S'
        elif 'T' in s_dt and '+08:00' in s_dt:  # '2015-05-20T13:29:35+08:00' rfc3339 标准格式 北京时间2015年05月20日13点29分35秒
            fmt = '%Y-%m-%dT%H:%M:%S%z'
            if has_dot:
                fmt = '%Y-%m-%dT%H:%M:%S.%f%z'
        elif has_dot:
            fmt = '%Y-%m-%d %H:%M:%S.%f'
        date_time = datetime.datetime.strptime(s_dt, fmt)

        # 添加 时分秒
        if add_hms and len(s_dt) <= 10:
            date_time += dt
        return date_time

    @staticmethod
    def str_to_time(s):
        """
        字符串 转 时间。
        s 格式：HH:MM:SS， 格式一点都不能错
        """
        if not s:
            return None
        if isinstance(s, datetime.time):
            return s
        return datetime.time.fromisoformat(s)

    @staticmethod
    def datetime_to_str(dt, fmt=DT_SHORT):
        """
        日期转 字符串
        :param dt:  要转换的日期 为 datetime 类型
        :param fmt: 转换格式    DT_SHORT = 1  # 年-月-日 y4-mm-dd
        :return:  成功返回转换后的字符串。 输入为None时，返回None
        """
        if dt is None or dt == '':
            return None
        fmt_str = '%Y-%m-%d'
        if fmt == DT_SHORT:
            fmt_str = '%Y-%m-%d'
        elif fmt == DT_LONG:
            fmt_str = '%Y-%m-%d %H:%M:%S'
        elif fmt == DT_LONG_LESS:
            fmt_str = '%Y-%m-%d %H:%M'
        elif fmt == DT_FULL:
            fmt_str = '%Y-%m-%d %H:%M:%S.%f'
        elif fmt == DT_FULL_EX:
            fmt_str = '%Y%m%d-%H%M%S-%f'
        elif fmt == DT_FULL_EX2:
            fmt_str = '%Y%m%d_%H%M%S_%f'
        elif fmt == DT_YM:
            fmt_str = '%Y-%m'
        elif fmt == DT_MD:
            fmt_str = '%m-%d'

        someday = datetime.datetime.strftime(dt, fmt_str)
        return someday

    @staticmethod
    def date_ymd(dt):
        return JdTools.datetime_to_str(dt, DT_SHORT)

    @staticmethod
    def date_to_str(dt):
        if isinstance(dt, datetime.datetime):
            val = datetime.datetime.strftime(dt, '%Y-%m-%d %H:%M:%S')
        elif isinstance(dt, datetime.date):
            val = datetime.date.strftime(dt, '%Y-%m-%d')
        elif isinstance(dt, datetime.time):
            val = datetime.time.strftime(dt, '%H:%M:%S')
        else:
            return dt
        return val

    @classmethod
    def int_to_date(cls, n, str_out=True):
        if isinstance(n, int) and 0 < n < 70000:
            d = datetime.datetime(1900, 1, 1) + datetime.timedelta(days=n-2)
            if str_out:
                d = cls.datetime_to_str(d)
            return d
        return None

    @staticmethod
    def time_to_datetime(t, dt=None):
        """ 日期字符串 转化为 datetime，例如 "09:00" 转为 "2019-11-18 09:00:00"，日期是当天 """
        n = dt or datetime.datetime.now()
        h = int(t.split(':')[0])
        m = int(t.split(':')[1])
        return datetime.datetime(year=n.year, month=n.month, day=n.day, hour=h, minute=m, second=0)

    @staticmethod
    def get_code(no=None, tag=None, school_no=None, date=None, hyphen=True,
                 no_width=3, school_width=4, full=False):
        """生成单据编号，包括学号，班级编号，收费单号等。

        :param no:              单据顺序号
        :param tag:             单据标识，例如学号为 XH， 收费单为 SF
        :param school_no:       分校编号，例如：0001,0002
        :param date:            日期，单据上附加日期信息。False，不带日期
        :param hyphen:          是否显示短横'-'
        :param no_width:        no 占用宽度，默认为3
        :param school_width:    分校编号占用宽度，默认为4
        :param full:            true: 日期格式为 年4月日时分秒。 false， 默认， 年2月日
        :return:        生成的编号, 例如 0001-XH-181008-001
        """
        fmt_no = '{:0%sd}' % (no_width or 3)
        fmt_sch = '{:0%sd}-' % (school_width or 4)

        fmt_date = '%y%m%d-' if not full else '%Y%m%d%H%M%S-'
        if date is None:
            date = datetime.datetime.today()
        data_str = datetime.datetime.strftime(date, fmt_date) if date is not False else ''
        sch_str = fmt_sch.format(int(school_no)) if school_no else ''
        tag_str = ('%s-' % tag) if tag else ''
        no_str = fmt_no.format(int(no)) if no is not None else ''
        code_str = '%s%s%s%s' % (sch_str, tag_str, data_str, no_str)
        if not hyphen:
            code_str = code_str.replace('-', '')
        return code_str

    @staticmethod
    def utc2local(utc_st):
        """
        UTC时间转本地时间（+8:00）
        :param utc_st:          utc datetime
        :return:                local datetime
        """
        now_stamp = time.time()
        local_time = datetime.datetime.fromtimestamp(now_stamp)
        utc_time = datetime.datetime.utcfromtimestamp(now_stamp)
        offset = local_time - utc_time
        local_st = utc_st + offset
        return local_st

    @staticmethod
    def local2utc(local_st):
        """
        本地时间转UTC时间（-8:00）
        :param local_st:        local datetime
        :return:                utc datetime
        """
        time_s = time.mktime(local_st.timetuple())
        utc_st = datetime.datetime.utcfromtimestamp(time_s)
        return utc_st

    @staticmethod
    def find_sub_str(string, sub, n):
        """查找字符串中子串第 n 次出现的位置索引
        :param string:   被查找字符串
        :param sub:     子串
        :param n:       第 n 次出现的位置， 从 1 开始
        :return:        子串 第 n 次出现位置的索引。若没有则返回 -1
        """
        arr = string.split(sub, n)
        if len(arr) <= n:
            return -1
        return len(string) - len(arr[-1]) - len(sub)

    @staticmethod
    def is_date(string, fmt=1):
        """
        判断输入字符串 是否 为 日期格式
        :param string:  要校验的字符串
        :param fmt:     日期类型， 2 DT_LONG(默认): yyyy-MM-dd hh:mm:ss，1 DT_SHORT : yyyy-MM-dd， 5 DT_YM: yyyy-MM
        :return: True 符合日期格式， False，不符合
        """
        if fmt == 1:  # yyyy-MM-dd
            pt = r"^[1-9]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1])$"
        elif fmt == 5:  # yyyy-MM
            pt = r"^[1-9]\d{3}-(0[1-9]|1[0-2])$"
        else:   # default 2  # yyyy-MM-dd hh:mm:ss
            pt = r'^[1-9]\d{3}-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1])\s+(20|21|22|23|[0-1]\d):[0-5]\d:[0-5]\d$'
        m = re.match(pt, string)
        # print(m)
        return m is not None

    @staticmethod
    def get_between_year_month(first_day, end_day):
        """
        根据开始日期和结束，获取中间月份
        :param first_day:      开始月份
        :param end_day:        结束年月
        :return:
        """
        date_arr = []
        months = (end_day.year - first_day.year) * 12 + end_day.month - first_day.month
        month_range = ['%s-%s' % (first_day.year + mon // 12, mon % 12 + 1)
                       for mon in range(first_day.month - 1, first_day.month + months)]
        for rec in month_range:
            year, month = JdTools.get_year_month(rec)
            # 将每个月末考勤日期存入数组
            if month != 12:
                end_day = datetime.datetime(year, month + 1, 1)
            else:
                end_day = datetime.datetime(year+1, 1, 1)
            end_day = end_day - datetime.timedelta(seconds=1)
            date_arr.append(end_day)
        return date_arr

    @staticmethod
    def is_birthday(days, date):
        # date 对应日期 是否进行生日提醒。 提前天数 为 days。
        today = datetime.date.today()
        end = today + datetime.timedelta(days=days)
        m = date.month
        d = date.day

        if today.month == end.month:    # 同一个月
            if m == today.month and today.day <= d <= end.day:
                return True
        elif today.month < end.month:   # 跨月
            if (m == today.month and d >= today.day) or (m == end.month and d <= end.day) \
                    or today.month < m < end.month:
                return True
        else:   # today.month > end.month   # 跨年情况
            if (m == today.month and d >= today.day) or (m == today.month, d >= today.day) \
                    or m > today.month or m < end.month:
                return True

        return False

    @staticmethod
    def get_birthday_diff(days, bd):
        # 比较今天距离生日日期间隔是否在 days 的天数内，如果是，返回间隔天数，否则返回-1
        y = datetime.date.today().year
        diff = -1
        if bd and isinstance(bd, (datetime.date, datetime.datetime)):
            d = (datetime.date(year=y, month=bd.month, day=bd.day) - datetime.date.today()).days
            if 0 <= d <= days:
                diff = d
        return diff

    @staticmethod
    def del_key(obj, key):
        try:
            del obj[key]
        except KeyError:
            pass

    @staticmethod
    def del_keys(obj, *args):
        for k in args:
            try:
                del obj[k]
            except KeyError:
                pass

    @staticmethod
    def list_remove(lst, e):
        try:
            lst.remove(e)
        except ValueError:
            pass

    @staticmethod
    def set_remove(obj, e):
        try:
            obj.remove(e)
        except KeyError:
            pass

    @staticmethod
    def excel_to_datetime(value):
        """
        excel 日期 (float) 转 日期格式:  40861.0 -> 2011-11-14 0:0:0

        使用 xlrd 从 excel 读取出日期后会变成一个 float 数值，比如  40861.5。
        其中整数部分表示 从 1900-1-1开始经过的天数。 1 -- 1900-1-1； 2 -- 1900-1-2
        小数部分表示 时间。24小数制。 0.5 -- 12:00:00, 0.15 -- 03:36:00
        24 * 60 * 60 = 86400(秒)

        :param value:   float, excel 读取的 日期值
        :return: 日期类型 datetime
        """
        if isinstance(value, str):
            _, value = JdTools.float(value)
        date = xlrd.xldate.xldate_as_datetime(value, 0)
        return date

    @staticmethod
    def name_phone(name, phone):
        # 获取学员 姓名、手机号 组成的 新名称。若 phone 为 None, ''，则不显示手机号
        _phone = (' (%s)' % JdTools.mask_phone(phone)) if phone else ''
        tip = '%s%s' % (name, _phone)
        return tip

    @staticmethod
    def mask_phone(phone):
        if phone:
            if len(phone) <= 4:
                _phone = phone
            else:
                _phone = phone[0:2] + '**' + phone[-4:]
        else:
            _phone = ''
        return _phone

    @staticmethod
    def check_id_card(id_card):
        """身份证验证"""
        area = {"11": "北京", "12": "天津", "13": "河北", "14": "山西", "15": "内蒙古",
                "21": "辽宁", "22": "吉林", "23": "黑龙江", "31": "上海", "32": "江苏",
                "33": "浙江", "34": "安徽", "35": "福建", "36": "江西", "37": "山东",
                "41": "河南", "42": "湖北", "43": "湖南", "44": "广东", "45": "广西",
                "46": "海南", "50": "重庆", "51": "四川", "52": "贵州", "53": "云南",
                "54": "西藏", "61": "陕西", "62": "甘肃", "63": "青海", "64": "宁夏",
                "65": "新疆", "71": "台湾", "81": "香港", "82": "澳门", "91": "国外"}
        id_card = str(id_card)
        id_card = id_card.strip()

        # 地区校验
        key = id_card[0: 2]  # 地区中的键是否存在
        if key not in area.keys():
            return False, '身份证地区非法！'
        # 15位身份号码检测

        if len(id_card) == 15:
            if (int(id_card[6:8]) + 1900) % 4 == 0 or \
                    ((int(id_card[6:8]) + 1900) % 100 == 0 and (int(id_card[6:8]) + 1900) % 4 == 0):
                re_str = '[1-9][0-9]{5}[0-9]{2}((01|03|05|07|08|10|12)(0[1-9]|[1-2][0-9]|3[0-1])|'
                re_str += '(04|06|09|11)(0[1-9]|[1-2][0-9]|30)|02(0[1-9]|[1-2][0-9]))[0-9]{3}$'
                erg = re.compile(re_str)  # // 测试出生日期的合法性
            else:
                re_str = '[1-9][0-9]{5}[0-9]{2}((01|03|05|07|08|10|12)(0[1-9]|[1-2][0-9]|3[0-1])|'
                re_str += '(04|06|09|11)(0[1-9]|[1-2][0-9]|30)|02(0[1-9]|1[0-9]|2[0-8]))[0-9]{3}$'
                erg = re.compile(re_str)  # // 测试出生日期的合法性
            if re.match(erg, id_card):
                return True, None
            else:
                return False, '身份证号码出生日期超出范围或含有非法字符！'
        # 18位身份号码检测
        elif len(id_card) == 18:
            # 出生日期的合法性检查
            # 闰年月日:((01|03|05|07|08|10|12)(0[1-9]|[1-2][0-9]|3[0-1])|(04|06|09|11)(0[1-9]|[1-2][0-9]|30)|02(0[1-9]|[1-2][0-9]))
            # 平年月日:((01|03|05|07|08|10|12)(0[1-9]|[1-2][0-9]|3[0-1])|(04|06|09|11)(0[1-9]|[1-2][0-9]|30)|02(0[1-9]|1[0-9]|2[0-8]))
            if int(id_card[6:10]) % 4 == 0 or (int(id_card[6:10]) % 100 == 0 and int(id_card[6:10]) % 4 == 0):
                re_str = '[1-9][0-9]{5}(19[0-9]{2}|20[0-9]{2})((01|03|05|07|08|10|12)(0[1-9]|[1-2][0-9]|3[0-1])'
                re_str += '|(04|06|09|11)(0[1-9]|[1-2][0-9]|30)|02(0[1-9]|[1-2][0-9]))[0-9]{3}[0-9Xx]$'
                erg = re.compile(re_str)  # // 闰年出生日期的合法性正则表达式
            else:
                re_str = '[1-9][0-9]{5}(19[0-9]{2}|20[0-9]{2})((01|03|05|07|08|10|12)(0[1-9]|[1-2][0-9]|3[0-1])'
                re_str += '|(04|06|09|11)(0[1-9]|[1-2][0-9]|30)|02(0[1-9]|1[0-9]|2[0-8]))[0-9]{3}[0-9Xx]$'
                erg = re.compile(re_str)  # // 平年出生日期的合法性正则表达式
            # 测试出生日期的合法性
            if re.match(erg, id_card):
                return True, None
            else:
                return False, '身份证号码出生日期超出范围或含有非法字符！'
        else:
            return False, '身份证号码位数不对！'

    @staticmethod
    def compare_time(time1, time2):
        """
        判断 两个时间段时间是否冲突
        :param time1:   时间段1, 例如 time1 = '08:00-09:00'
        :param time2:   时间段2，例如 time2 = '08:30-09:30'
        :return: 冲突返回 False, 不冲突，返回 True

        例：time1 = '08:00-09:00'
            time2 = '08:30-09:30'
            返回False
        """
        arr1 = time1.split('-')
        h1 = arr1[0].split(':')  # 开始时间
        h2 = arr1[1].split(':')  # 结束时间
        begin_time1 = int(h1[0]) * 100 + int(h1[1])  # 开始时间整数形式
        end_time1 = int(h2[0]) * 100 + int(h2[1])  # 结束时间整数形式
        arr2 = time2.split('-')
        h1 = arr2[0].split(':')  # 开始时间
        h2 = arr2[1].split(':')  # 结束时间
        begin_time2 = int(h1[0]) * 100 + int(h1[1])  # 开始时间整数形式
        end_time2 = int(h2[0]) * 100 + int(h2[1])  # 结束时间整数形式
        if end_time2 > begin_time1 and end_time1 > begin_time2:
            return False
        return True

    @staticmethod
    def decimal(number, result=False, pt=2):
        """将number数值转换为decimal类型，支持数值类型：int,str,float
            改为 四舍五入方式 ROUND_HALF_UP。 默认为 ROUND_HALF_EVEN 四舍六入，五平均（前一位是偶数时进位）
            result = True, 返回  number, 成功标识
            result = False, 返回  number
        """
        success = True
        exp = Decimal('0.' + pt * '0')
        if number == '' or number is None:      # 容错处理： None, ''(空串), 返回 0
            number = 0
        elif isinstance(number, float):
            number = Decimal.from_float(number).quantize(exp, rounding=ROUND_HALF_UP)
        elif isinstance(number, int):
            number = Decimal(number)
        elif isinstance(number, str):
            try:
                number = Decimal(number).quantize(exp, rounding=ROUND_HALF_UP)
            except (InvalidOperation, DecimalException):
                success = False
        elif isinstance(number, Decimal):
            number = number.quantize(exp, rounding=ROUND_HALF_UP)
        else:
            success = False

        if result:  # 需要返回是否转换成功结果时
            return number, success
        else:
            return number

    @staticmethod
    def decimal_check(number, pt=2):
        """对 JdTools.decimal 返回值进行 颠倒， 返回  (判断结果，number)"""
        n, check = JdTools.decimal(number, True, pt)
        return check, n

    @staticmethod
    def create_qrcode(content, center_img=None):
        """创建二维码
        content     str     创建二维码的内容
        center_img  str     中心图片，可选
        """
        qr = qrcode.QRCode(version=5, error_correction=qrcode.constants.ERROR_CORRECT_H, box_size=8, border=4)
        qr.add_data(content)
        qr.make(fit=True)

        img = qr.make_image()
        img = img.convert("RGBA")
        if center_img and os.path.isfile(center_img):
            icon = Image.open(center_img)  # 这里是二维码中心的图片
            img_w, img_h = img.size
            factor = 4
            size_w = int(img_w / factor)
            size_h = int(img_h / factor)

            icon_w, icon_h = icon.size
            if icon_w > size_w:
                icon_w = size_w
            if icon_h > size_h:
                icon_h = size_h
            icon = icon.resize((icon_w, icon_h), Image.ANTIALIAS)

            # 先添加白色框，比 logo 图标 大
            white_w = icon_w + 20
            white_h = icon_h + 20
            w = int((img_w - white_w) / 2)
            h = int((img_h - white_h) / 2)
            new_image = Image.new('RGBA', (white_w, white_h), (255, 255, 255))  # 生成白色图像
            img.paste(new_image, (w, h), new_image)

            w = int((img_w - icon_w) / 2)
            h = int((img_h - icon_h) / 2)
            icon = icon.convert("RGBA")
            img.paste(icon, (w, h), icon)
        return img

    @staticmethod
    def file_type(suffix, _type=0):
        """用于判断文件后缀是否在允许范围内
        suffix:  文件后缀名 例： jpg
        _type: 文件类型 0-图片 1-视频 2-待增加 默认为0
        """
        img_type = ['png', 'gif', 'jpg', 'jpeg', 'ico', 'bmp']  # 图片格式
        video = ['mpg', 'mpeg', 'avi', 'rm', 'rmvb', 'mov', 'wmv', 'asf', 'dat', 'mp4']  # 视频格式
        if _type == 0:  # 判断图片后缀是否在允许范围内
            if suffix not in img_type:
                return False
        elif _type == 1:  # 视频
            if suffix not in video:
                return False
        else:  # 其他
            return False
        return True

    @staticmethod
    def this_week(page=0, today=None):
        """通过星期，获取一周范围内日期 从星期一到星期日
            page:  翻页 默认为 0 本页
                   正数 例：1 下一周 2 下下周
                   负数 例：-1 上一周 2 上上周
            today: 当前日期 预留字段
        """
        if today is None:
            today = datetime.date.today()
        dates = {}  # 当前周 周一到周日日期集合
        week_day = today.weekday() + 1  # 当前星期
        one_day = datetime.timedelta(days=1)  # 一天
        if page > 0:  # 当前周之后
            today = today + ((7 * page) * one_day)
        elif page < 0:  # 当前周之前
            today = today + ((7 * page) * one_day)
        for i in range(1, 8):
            date = today - ((week_day - i) * one_day)
            dates[i] = date
        return dates

    @staticmethod
    def this_week2(page=0, today=None, num=None):
        """通过星期，获取一周范围内日期 从当天开始连续7天之内的星期
            page:  翻页 默认为 0 本页
                   正数 例：1 下一周 2 下下周
                   负数 例：-1 上一周 2 上上周
            today: 当前日期 预留字段
        """
        if today is None:
            today = datetime.date.today()
        dates = {}  # 当前周 周一到周日日期集合
        one_day = datetime.timedelta(days=1)  # 一天
        if page > 0:  # 当前周之后
            today = today + ((num * page) * one_day)
        elif page < 0:  # 当前周之前
            today = today + ((num * page) * one_day)
        for i in range(1, 8):
            date = today + (i * one_day)
            weekday = date.weekday() + 1
            dates[weekday] = date
        return dates

    @staticmethod
    def assign(out_dict, **kwargs):
        """对字典 out_dict 赋值， 按 kwargs 中 k， v。示例：

        d = {m: 3}
        assign(d, times=1, fee=100, m=2, n=None)
            d  => {times: 1, fee: 100, m:5, n: 0}
        """
        for k, v in kwargs.items():
            out_dict[k] = (v or 0) + (out_dict.get(k) or 0)

    @staticmethod
    def assign_ex(out_dict, **kwargs):
        """对字典 out_dict 赋值， 按 kwargs 中 k， v。示例：
        原来字典中 None值的key，若未累加，保持 None 值不变

        d = {m: 3, x: None}
        assign(d, times=1, fee=100, m=2, x=0, y=None)
            d  => {times: 1, fee: 100, m:5, x: None}
        """
        for k, v in kwargs.items():
            if not v:  # 输入为 0 或者 None，不做计算
                continue
            out_dict[k] = (v or 0) + (out_dict.get(k) or 0)

    @staticmethod
    def nested(out_dict, *args):
        """字典 out_dict 嵌套赋值， 按 args 中 v 逐级嵌套。 示例：

        d = {}, class_id = 10, student_id = 900
        nested(d, class_id, student_id)
        d  => {10: {900: {} } }
        """
        d = out_dict
        for v in args:
            if v not in d:
                d[v] = {}
            d = d[v]

    @staticmethod
    def calculate_age(born):
        # 根据生日获得其年龄，born 参数为 datetime.date 类型
        # 周岁年龄 = 当前年份 - 出生日期年份。周岁年龄又称实足年龄，指从出生到计算时为止，共经历的周年数或生日数。
        #
        # 例如，1990年7月1日零时进行人口普查登记，一个1989年12月15日出生的婴儿，按虚岁计算是2岁，实际刚刚6个多月，
        # 还未过一次生日，按周岁计算应为不满1周岁，即0岁。
        # 周岁年龄比虚岁年龄常常小1～2岁，它是人口统计中常用的年龄计算方法。周岁—出生时为0岁，每过一个公历生日长1岁。
        today = datetime.date.today()
        try:
            birthday = born.replace(year=today.year)
        except ValueError:
            # raised when birth date is February 29
            # and the current year is not a leap year
            birthday = born.replace(year=today.year, day=born.day - 1)
        if birthday > today:
            return today.year - born.year - 1
        else:
            return today.year - born.year

    @staticmethod
    def int2time_str(n):
        """int转时间字符串，如800 -> 8:00,  未校验 分钟超过 59分的异常情况 """
        if isinstance(n, int) and 0 <= n < 2400:
            return '%02d:%02d' % (n // 100, n % 100)
        return None

    @staticmethod
    def time_str2int(s):
        """时间字符串转int，如8:00转800"""
        if isinstance(s, str) and ':' in s:
            list_ = s.split(':')
            try:
                h = int(list_[0])
                m = int(list_[1])
            except ValueError:
                return None
            if 0 <= h < 24 and 0 <= m <= 59:
                output = h * 100 + m
                return output
        return None

    @staticmethod
    def int2datetime(n, date: datetime.datetime = None):
        """int转datetime类型，日期为当天，如800 -> datetime.datetime(2021, 11, 5, 8, 0)"""
        if isinstance(n, int) and 0 <= n < 2400:
            now = date if date else datetime.datetime.now()
            h = n // 100
            m = n % 100
            output = datetime.datetime(now.year, now.month, now.day, h, m)
            return output

        return None

    @staticmethod
    def obj_from_key(obj: dict, key: str):
        """ 从 对象 obj 中取出 key 所对应的对象
            例如 key = 'show' 返回  obj['show']
            例如 key = 'program.show' 返回  obj['program']['show']
        """
        if key == '':
            return obj
        arr = key.split('.')
        o = obj

        for i in arr:
            o = o.get(i)
        return o

    @staticmethod
    def fen2yuan(s):
        """分转为元，微信支付专用
        :param s: 数字字符串，微信返回，单位为“分”
        :return yuan: Decimal类型，单位“元”
        """
        _, fen = JdTools.int(s)
        yuan = JdTools.decimal(fen / 100)
        return yuan

    @staticmethod
    def yuan2fen(yuan) -> int:
        """元 转 分"""
        return int(yuan * 100)

    @staticmethod
    def copy_dict(obj: dict, *args):
        # 安全拷贝 字典。将 字典 obj 中 选中的 key 生成新的字典
        out = {}
        d = obj
        for v in args:
            if v in d:
                out[v] = d[v]
        return out


class JdDate(object):
    """日期相关工具函数"""
    days = [31, ]

    @staticmethod
    def get_date(str_date, day=1):
        """获取 datetime.date 日期， str_date 格式为 年-月 or 年-月-日  字符串"""
        y, m = JdTools.get_year_month(str_date)
        start_date = datetime.date(y, m, day)
        return start_date

    @staticmethod
    def get_datetime(str_date, day=1):
        """获取 datetime.datetime 日期， str_date 格式为 年-月 or 年-月-日 字符串"""
        y, m = JdTools.get_year_month(str_date)
        start_date = datetime.datetime(y, m, day)
        return start_date

    @staticmethod
    def next_day(date):
        """获取 datetime.datetime 日期 的后一天， 返回 datetime """
        n = date + datetime.timedelta(days=1)
        m = datetime.datetime(year=n.year, month=n.month, day=n.day)
        return m

    @staticmethod
    def last_day_of_month(any_day):
        """
        获取一个月中的最后一天
        :param any_day: 任意日期    datetime.datetime, or datetime.date
        :return: string
        """
        next_month = any_day.replace(day=28) + datetime.timedelta(days=4)  # 确保到下一个月
        d = next_month - datetime.timedelta(days=next_month.day)
        if isinstance(d, datetime.datetime):
            d += datetime.timedelta(hours=23-d.hour, minutes=59-d.minute, seconds=59-d.second)
        return d

    @staticmethod
    def last_time_of_day(date: datetime.datetime) -> datetime.datetime:
        """获取某天的 23:59:59"""
        return datetime.datetime(date.year, date.month, date.day, 23, 59, 59)

    @staticmethod
    def begin_time_of_day(date: datetime.datetime) -> datetime.datetime:
        """获取某天的 00:00:00"""
        return datetime.datetime(date.year, date.month, date.day, 00, 00, 00)

    @staticmethod
    def year_month(date):
        """从 日期 对象返回 年-月 字符串 """
        if isinstance(date, (datetime.datetime, datetime.date)):
            month = '%d-%02d' % (date.year, date.month)
            return month
        return '<日期格式错误>'

    @staticmethod
    def prev_month(date):
        """ 1. 获取 月份字符串 2019-08 的前一个月： 2019-07
            2. date 为 datetime时， 返回值也是 datetime
        """
        return JdDate.month_dist(date, -1)

    @staticmethod
    def month_dist(date, month_num):
        """ 1. 获取 月份字符串 2019-08 的前/后 month_num 月： -1, 2019-07; 2, 2019-10
               month_num 正数，向后 偏移月份； 负数，向前
            2. date 为 datetime时， 返回值也是 datetime
        """
        if isinstance(date, str):
            y, m = JdTools.get_year_month(date)
        else:
            y = date.year
            m = date.month
        m += month_num - 1
        y += m // 12    # 计算 年份偏移
        m = m % 12 + 1

        if isinstance(date, str):
            month = '%d-%02d' % (y, m)
        else:
            month = datetime.datetime(year=y, month=m, day=1)
        return month

    @staticmethod
    def check_date(param, field, name):
        """
        校验日期格式是否正确。
        :param param:       被校验 对象 dict
        :param field:       字段，dict[field]必须存在
        :param name:        field 的 中文名称。用于错误提示
        :return:  {code: 错误码， msg： 错误提示}
        """
        # ERR_DATE_FMT = 200
        # ERR_DATE_FMT_MSG = '“%s”格式错误【%s】，期望 年-月-日！'
        tip = '“%s”格式错误【%s】，期望 年-月-日！'
        date = param.get(field)
        if isinstance(date, int):   # xlrd 读取日期格式的单元格。xls格式为 2， xlsx格式为 3（在读取时，已经转为日期）
            if date > 2958465 or date < 1:   # 日期范围 1990/01/01 - 9999/12/31
                return {'code': 200, 'msg': '输入日期【{}】不在正常范围内'.format(date)}
            d1 = xlrd.xldate.xldate_as_datetime(date, 0)
            date = d1
        try:
            param[field] = JdTools.str_to_datetime(date)
        except (ValueError, TypeError):
            return {'code': 200, 'msg': tip % (name, date)}
        return {'code': 0, 'msg': 'ok'}

    @staticmethod
    def verify_date(obj, **kwargs):
        """
        校验日期格式是否正确。
        :param obj:         被校验 对象, dict
        :param kwargs:      被校验字段，dict, {字段名称='中文字段名称'}
        :return:    {code: 错误码， msg： 错误提示}
        """
        tip = '“%s”格式错误【%s】，期望 年-月-日！'
        for k, v in kwargs.items():
            dt = obj.get(k)
            try:
                obj[k] = JdTools.str_to_datetime(dt)
            except (ValueError, TypeError):
                return {'code': 200, 'msg': tip % (v, dt)}
        return {'code': 0, 'msg': 'ok'}

    @staticmethod
    def month_num(month, num):
        # 返回 月，月数 对应 字符串。 month: str, 起始月，例如 '2019-10', num: int 月数
        if num == 1:
            return month
        else:
            end = JdDate.month_dist(month, num)
            return '%s到%s共%s月' % (month, end, num)

    @staticmethod
    def start_end_time(tm, val=None):
        try:
            d1, d2 = start_end_time(tm, val)
        except (AttributeError, ValueError, TypeError):
            return {'code': 600, 'msg': f'日期格式错误【{tm} {val}】'}

        if d1.date() == d2.date():  # 今天或者昨天的时间段
            date_from_to = '%s' % (JdTools.date_ymd(d1))
        else:
            date_from_to = '从 %s 至 %s' % (JdTools.date_ymd(d1), JdTools.date_ymd(d2))
        return {'code': 0, 'msg': 'ok', 'd1': d1, 'd2': d2, 'date_from_to': date_from_to}

    @staticmethod
    def is_leap_year(y):
        # 判断输入年 y 是否为公历闰年（满足以下任意条件： 1. 能被400整除，2. 能被4整除，不能被100整除）
        if y % 400 == 0 or (y % 4 == 0 and y % 100 != 0):
            return True
        return False

    @staticmethod
    def days_distant(m, d):
        # 判断 公历日期 月-日 (m-d) 和 “今天”的日期差。
        # 将 m-d 转换为 今年的日期。注意闰年
        today = datetime.date.today()
        y = today.year
        # 公历闰年 2-29
        if m == 2 and d == 29 and not JdDate.is_leap_year(y):
            d = (datetime.date(year=y, month=3, day=1) - today).days  # 公历 2-29日生日，在平年的3-1日作为生日
        else:
            d = (datetime.date(year=y, month=m, day=d) - today).days
        return d

    @staticmethod
    def get_dates_by_times(start_day, end_day):
        """根据开始日期、结束日期返回这段时间里所有天的集合"""
        result = []
        date_start = datetime.datetime.strptime(start_day, '%Y-%m-%d')
        date_end = datetime.datetime.strptime(end_day, '%Y-%m-%d')
        result.append(date_start.strftime('%Y-%m-%d'))
        while date_start < date_end:
            date_start += datetime.timedelta(days=1)
            result.append(date_start.strftime('%Y-%m-%d'))
        return result

    @staticmethod
    def get_hours_list():
        """返回最近24小时时间字符串列表，时间格式为 %Y-%m-%d %H"""
        now = datetime.datetime.now()
        value = []
        begin = 0
        while begin <= now.hour:
            if begin < 10:
                begin_str = '0' + str(begin)
            else:
                begin_str = str(begin)
            value.append(now.strftime('%Y-%m-%d' + ' ' + begin_str))
            begin += 1
        max_hour = 23
        s = 24 - len(value)
        if s:
            yesterday = (now.today() + datetime.timedelta(days=-1)).strftime("%Y-%m-%d")
            while s:
                if max_hour < 10:
                    begin_str = '0' + str(max_hour)
                else:
                    begin_str = str(max_hour)
                value.insert(0, yesterday + ' ' + begin_str)
                max_hour -= 1
                s -= 1
        return value

    @staticmethod
    def equal(a, b):
        # 日期是否相同， 忽略 毫秒
        if isinstance(a, datetime.datetime) and isinstance(b, datetime.datetime):
            return a.year == b.year and a.month == b.month and a.day == b.day \
                   and a.hour == b.hour and a.minute == b.minute and a.second == b.second

        return False


class JdImage(object):
    """图片处理公共类"""

    @staticmethod
    def resize(infile, outfile=None, width=1280):
        """
        修改图片尺寸
        :param infile:  原始图片， 路径+文件，必须存在
        :param outfile: 修改后的图片，不填写，则覆盖原始图片
        :param width:   修改后的宽度，默认 1280
        :return: 无
        """
        im = Image.open(infile)
        (x, y) = im.size  # read image size
        _, width = JdTools.int(width, 1280)
        if x <= width:  # 原始图片小于调整后的 图片宽度，则不做调整
            return
        x_s = width     # define standard width
        y_s = int(y * x_s / x)  # calc height based on standard width
        out = im.resize((x_s, y_s), Image.ANTIALIAS)  # resize image with high-quality
        out.save(outfile or infile)


class JdFmt(object):
    """格式化类"""

    @staticmethod
    def has_power(val, has_power=True):
        """是否有权限，显示 val， 若 has_power==True，则显示 val，否则显示 **** """
        return val if has_power else ('****' if val else 0)

    @staticmethod
    def discount_rate(val):
        # 转换折扣率  val float 小数点4位，转 百分比，保留小数点后2位
        v = ''
        if val:
            v = '%s' % JdTools.get_float2(val * 100) + '%'
        return v


class JdList(object):
    @staticmethod
    def index(lst, el):
        try:
            idx = lst.index(el)
        except ValueError:
            idx = -1
        return idx


class JdPath(object):
    @staticmethod
    def slash(url):
        url_new = url.replace('\\', '/')  # 将 反斜杠替换为 斜杠， 防止 url 打不开的问题。
        return url_new


def name_list(class_list, key='class_name'):
    """ 提取列表元素中的班级名称，过滤空值和重复值，返回字符串"""
    classes = set()
    for r in class_list:
        class_name = r.get(key)
        if class_name:
            classes.add(class_name)
    return '，'.join(list(classes))


def sum_ticket(class_receipt, arr='ticket', key='fee'):
    """提取列表元素 some_list(数组中数组) 中的现金券总面值"""
    s = 0
    for r in class_receipt:
        s += sum_list(r[arr], key)
    return s


def sum_list(some_list, *key):
    """提取列表元素 some_list 中的现金券总面值"""
    s = 0
    for r in some_list:
        for kk in key:
            s += r.get(kk) or 0
    return s


class Singleton(type):
    # 单例模式，用法  class YourClass(metaclass=Singleton):
    def __init__(cls, *args, **kwargs):
        cls.instance = None
        super().__init__(*args, **kwargs)

    def __call__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)
        return cls.instance


# 通过装饰器实现
def singleton(cls):
    # 创建一个字典用来保存类的实例对象
    _instance = {}

    def _singleton(*args, **kwargs):
        # 先判断这个类有没有对象
        if cls not in _instance:
            _instance[cls] = cls(*args, **kwargs)  # 创建一个对象,并保存到字典当中
        # 将实例对象返回
        return _instance[cls]

    return _singleton


def humanize_date_delta(start, end):
    """
    将时间差转换为友好格式，如“7天”，“3个月”，“2年”， end > start，否则返回 空''
    :param start:     起始时间，datetime.datetime 或 datetime.date
    :param end:       结束时间，datetime.datetime 或 datetime.date
    :return:          转换结果，string
    """
    if not isinstance(start, datetime.date) or not isinstance(end, datetime.date):
        return ''

    if isinstance(start, datetime.datetime):
        start = start.date()
    if isinstance(end, datetime.datetime):
        end = end.date()

    days = (end - start).days
    if days <= 0:  # 结束时间早于或等于开始时间的情况
        return ''

    if days < 30:
        return f'{days}天'
    elif 30 <= days < 365:
        months = days // 30
        return f'{months}个月'
    else:  # days >= 365
        years = days // 365
        return f'{years}年'



# -*- coding:utf-8 -*-
# 文件:  errcode.py

ERR_DATE_FMT = 200
ERR_DATE_FMT_MSG = '“%s”格式错误【%s】，期望 年-月-日！'
ERR_TIME_FMT = 201
ERR_VERIFY_NEED_PARAM = 500     # 缺少参数
ERR_VERIFY_VAL_EXCEED = 501     # 数值越限
ERR_PUB_LEN = 700
ERR_PUB_LEN_MSG = '“%s”不能超过“%s”字！'
ERR_PUB_NUM = 706
ERR_PUB_NUM_MSG = '【%s】格式错误，请输入数值！'
ERR_PUB_NUM_INT = 707
ERR_PUB_NUM_INT_MSG = '【%s】格式错误，请输入整数！'
ERR_PUB_NUM_PT2 = 708
ERR_PUB_NUM_PT2_MSG = '【%s】最多两位小数！'
ERR_PUB_NEED = 709
ERR_PUB_NEED_MSG = '请输入【%s】!'
ERR_PUB_GT_EQ_0 = 713
ERR_PUB_GT_EQ_0_MSG = '“%s”需大于或等于0！'
ERR_PUB_CUSTOMIZED_ERR = 819
ERR_PUB_CUSTOMIZED_ERR_MSG = '自定义错误提示，通过ppt设置'
ERR_VERIFY_ERROR = 900          # 校验出错
ERR_VERIFY_WHO = 901            # 未知的校验类型
ERR_SYNTAX = 902
ERR_SYNTAX_MSG = '语法错误！[%s]'



# -*- coding: utf-8 -*-
import os
from .tools import Singleton, JdPath, JdTools
from .oper_file import copy_file, move_file


__version__ = '1.0.6'
__all__ = ('Avatar', 'AvatarDirMgr')

"""
文件上传， 将临时上传的文件移动 / 目录重命名为 正式目录
只支持单文件操作

用法示例：
1. 先注册模块
    ad = AvatarDirMgr()
    ad.basedir = os.path.join(basedir, 'app/static')
    ad.tmpdir = os.path.join(basedir, 'app/static/tmp')
    logger.debug(f"basedir: [{ad.basedir}], tmpdir: [{ad.tmpdir}]")

    ad.register_module('user', 'head')  # 产生的url为： /head/user/{id}/filename
    ad.register_module('student')       # 产生的url为： /student/{id}/filename
"""

# location 取值定义
LOCAL_SERVER = 1    # 本地服务器
PIC_SERVER = 2      # 图片服务器


class AvatarDirMgr(metaclass=Singleton):
    _module = {}    # 模块
    _basedir = ''   # 文件基准目录。 静态页面所在地址
    _tmp_dir = ''   # 公共临时文件相对目录， 从 _basedir 目录开始写

    @property
    def basedir(self):
        return self._basedir

    @basedir.setter
    def basedir(self, basedir):
        if not os.path.isdir(basedir):
            raise Exception(f'basedir path not exist! [{basedir}]')
        self._basedir = basedir

    @property
    def tmpdir(self):
        return self._tmp_dir

    @tmpdir.setter
    def tmpdir(self, tmp_dir):
        self._tmp_dir = tmp_dir

    def register_module(self, module: str, m_dir: str = '', absolute=False, prefix=''):
        """
        注册模块 及存放路径
        :param module:  模块名称
        :param m_dir:   存放路径，基准路径后面的路径，比如 '/img/head/', 通过 域名 + 该路径 可以访问到文件
        :param absolute:  绝对目录， False-否， 在 m_dir 后添加【模块名】做目录， True-是，不添加 模块名目录
        :param prefix:  id 目录前缀， 比如 company， id为5， 生成的目录为 company_5
        :return:
        """
        if not module:
            raise Exception('[module] is not assigned!')
        # m_dir 以 '/' 开头时， 去掉 '/'
        self._module[module] = {'dir': m_dir[1:] if m_dir and m_dir[0] == '/' else (m_dir or ''),
                                'absolute': absolute, 'prefix': prefix or ''}

    def allowed(self, module: str) -> bool:
        """
        判断模块是否注册过
        :param module:  模块名称
        :return:    bool， True-模块注册过， False-模块未注册
        """
        return module in self._module

    def get_dir(self, module: str) -> tuple:
        """
        获取模块的存放路径
        :param module:  模块名称
        :return:    (模块存放路径, absolute, prefix)，若模块未注册，返回 空 字符串
        """
        m = self._module.get(module)
        if m:
            return m['dir'], m['absolute'], m['prefix']
        return '', False, ''


class Avatar(object):
    """
    图片文件处理
    负责 某模块 图片 位置存储， 临时文件存储， 临时文件转到正式目录
    """

    def __init__(self, module, module_id=0, relative_dir=False):
        self.__module = module
        self.__module_id = module_id

        self.__url = ''     # 图片文件 url
        self._mgr = AvatarDirMgr()
        if not self._mgr.allowed(module):
            raise Exception(f'module [{module}] not registered!')
        self._basedir = self._mgr.basedir
        self._dir, self._abs, self._prefix = self._mgr.get_dir(module)
        self._tmpdir = self._mgr.tmpdir
        self._relative_dir = relative_dir   # 临时目录使用相对目录， True-是， False-否，使用公共临时目录
        
    def file_path(self, url):
        """
        获取 url 对应的服务器文件路径
        """
        if not url:
            return ''
        fp = os.path.join(self._basedir, url[1:] if url[0] == '/' else url)
        return JdPath.slash(fp)

    def _inner_path(self, some_id=None):
        f"""
        内部路径， 模块注册路径+之后的目录， 比如注册目录为 img/head,  模块为 'user', 则返回的内部路径为
        img/head/user/{some_id}
        若 模块ID为0，新增时，返回的路径为
        tmp_dir/del_user_年月日_时分秒_6位随机字符
        :param some_id:     模块ID
        :return:        内部路径
        """
        module = self.__module
        module_id = some_id or self.__module_id
        if module_id == 0:
            path = f'del_{module}_' + JdTools.get_date()
            upper_dir = self._dir if self._relative_dir else self._tmpdir
        else:
            if self._abs:   # True-是，不添加 模块名目录
                path = f"{self._prefix}{module_id}"
            else:
                path = f'{module}/{self._prefix}%s' % module_id  # 内部路径是 模块名/数据id
            upper_dir = self._dir
        return os.path.join(upper_dir, path)

    def get_module_path(self, some_id=None):
        # 获取模块所在路径， 文件真实路径
        module_id = some_id or self.__module_id
        directory = self._inner_path(module_id)
        dd = os.path.join(self._basedir, directory)
        d2 = JdPath.slash(dd)
        return d2

    def _mk_dir(self, some_id=None, create_dir=True):
        # 创建目录
        module_id = some_id or self.__module_id
        directory = self._inner_path(module_id)
        dd = os.path.join(self._basedir, directory)
        if create_dir and not os.path.exists(dd):  # 当路径不存在时
            os.makedirs(dd)  # 创建路径
        return dd

    def new_file(self, new_file, create_dir=True, some_id=None):
        """返回新文件路径全称
        new_file:       文件名，不含路径
        create_dir:     是否创建新文件所在目录
        some_id:        记录ID，不填时使用模块ID
        """
        module_id = some_id or self.__module_id
        dd = self._mk_dir(module_id, create_dir)
        fp = os.path.join(dd, new_file)

        url = fp.split(self._basedir)[-1]
        self.__url = JdPath.slash(url)  # 修改斜杠，避免url打不开
        return fp

    @property
    def url(self):
        return self.__url

    @staticmethod
    def rm(fn):
        """删除图片文件  fn: 头像 url """

        if not fn or fn in ['/img/avatar.jpg', '/img/item_default.png', '/img/logo.png']:
            return

        head = fn[1:] if fn[0] == '/' else fn
        old_img = os.path.join(AvatarDirMgr().basedir, head)
        if os.path.isfile(old_img):
            try:
                os.remove(old_img)
            except FileNotFoundError:
                pass

    @classmethod
    def basename(cls, filepath):
        """获取文件名， filepath 为路径+文件名"""
        return os.path.basename(filepath)

    @classmethod
    def copy_file_to(cls, module, module_id, old_url):
        """拷贝文件到模块所在目录
        old_url:    原文件所在目录
        """
        av = Avatar(module, module_id)
        filename = av.basename(old_url)

        old_fp = av.file_path(old_url)
        new_fp = av.new_file(filename)
        copy_file(old_fp, new_fp)
        return av.url
    
    @classmethod
    def move_to_normal(cls, module, url, some_id):
        """
        将临时图片 移动到 正式目录， 不保存数据库。 返回 正式 url 地址
        重命名临时目录 --> 正式目录(正式目录不能提前创建) 或者 将 临时目录文件拷贝到正式目录后，再删除 临时目录
        :param module:      模块名称
        :param url:         临时文件url
        :param some_id:     模块ID
        :return:
        """
        image = url
        some_id = some_id
        ret_url = url
        if image:
            av = Avatar(module, some_id)
            old_file = av.file_path(image)
        
            old_path = os.path.dirname(old_file)  # 获取文件所在路径
            new_path = av.get_module_path()
            if old_path == new_path:
                return ret_url
            fn = os.path.basename(old_file)
            new_file = av.new_file(fn, create_dir=False)
            ret_url = av.url
        
            try:
                os.renames(old_path, new_path)  # new_path 存在时，报错跳转到 FileExistsError
            except FileNotFoundError:
                pass
            except FileExistsError:
                move_file(old_file, new_file)
                if not os.listdir(old_path):  # 目录没有文件
                    os.rmdir(old_path)  # 删除目录
    
        return ret_url

    @classmethod
    def move_to_normal_db(cls, module, tb, avatar_key):
        """
        将临时图片 移动到 正式目录。 同时修改表 的字段，改为新的 图片文件url
        重命名临时目录 --> 正式目录(正式目录不能提前创建) 或者 将 临时目录文件拷贝到正式目录后，再删除 临时目录
        :param module:      模块名称
        :param tb:          表对象
        :param avatar_key:  表中的字段名
        :return:
        """
        image = getattr(tb, avatar_key)
        some_id = getattr(tb, 'id')
        ret_url = cls.move_to_normal(module, image, some_id)
        setattr(tb, avatar_key, ret_url)
        return ret_url
        


# -*- coding: utf8 -*-
""" 读取程序启动所需的配置，从配置文件 startup.ini 中读取配置。

用法：
cfg = Cfg("test.ini")     # 实例化 启动配置文件 读取对象
print(cfg.sections())

===============================================================================
Author:     WXG
Date:       2019-07-17

DATE        AUTHOR  DESCRIPTION
----------  ------  -----------------------------------------------------------------
2022-07-08  WXG     cfg 新增获取字符串类型的函数 str_val , 版本改为 2.3
2022-07-12  WXG     cfg 新增获取配置文件 段名称列表函数 sections , 版本改为 2.5
2022-07-23  WXG     redis 配置 移动到段[redis]中 , 版本改为 2.7
"""
from .parser_config import JdIniReader, NO_SECTION_OR_KEY

__all__ = ['Cfg', 'STARTUP_SECTION_TASKS', 'STARTUP_SECTION_DB']
__version__ = '2.9.3'


STARTUP_SECTION_RUN = 'run'
STARTUP_SECTION_REDIS = 'redis'             # redis 配置
STARTUP_SECTION_DB = 'db'                   # 主数据库
STARTUP_SECTION_FACE = 'face_recognize'     # 人脸识别
STARTUP_SECTION_CARD = 'card_recognize'     # 读卡器
STARTUP_SECTION_GZH = 'gzh'                 # 微信公众号相关配置
STARTUP_SECTION_TASKS = 'tasks'             # 定时任务 配置段


class FaceRecognize(object):
    """人脸识别参数"""
    def __init__(self, host, port, output):
        self._host = host
        self._port = port
        self._output = output

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def output(self):
        return self._output


class Cfg(object):
    """程序启动后读取配置文件， 如配置文件/Section/key不存在，对应值返回 0 ****"""

    def __init__(self, path_filename):
        # 读取默认配置文件
        try:
            self.__startup_cfg = JdIniReader(path_filename)
        except IOError:
            self.__startup_cfg = None

        # 是否 腾讯 token 主控服务器    0 否， 1 是
        self.__is_token_server = self.__get_v(STARTUP_SECTION_RUN, 'is_token_server')
        # 是否创建微信公众号自定义菜单    0 否， 1 是
        self.__is_create_menu = self.__get_v(STARTUP_SECTION_RUN, 'create_wx_menu')
        self.__sandbox = self.__get_v(STARTUP_SECTION_RUN, 'sandbox')
        self.__wx_test = self.__get_v(STARTUP_SECTION_RUN, 'wx_test')
        self.__user = self.__get_v(STARTUP_SECTION_DB, 'user')
        self.__password = self.__get_v(STARTUP_SECTION_DB, 'pwd')
        self.__database = self.__get_v(STARTUP_SECTION_DB, 'database')
        v = self.__get_v(STARTUP_SECTION_DB, 'host')
        self.__host = 'localhost' if v == 0 else v
        self.__port = self.__get_v(STARTUP_SECTION_DB, 'port', 3306)
        self.__is_mysql = self.__get_v(STARTUP_SECTION_DB, 'use_mysql', 1)
        self.__start_face_server = self.__get_v(STARTUP_SECTION_RUN, 'start_face_server')
        v = self.__get_v(STARTUP_SECTION_FACE, 'host')
        face_host = 'localhost' if v == 0 else v
        self.face = FaceRecognize(face_host,
                                  self.__get_v(STARTUP_SECTION_FACE, 'port', 9999),
                                  self.__get_v(STARTUP_SECTION_FACE, 'output'))

        self.__start_card_server = self.__get_v(STARTUP_SECTION_RUN, 'start_card_server')
        v = self.__get_v(STARTUP_SECTION_CARD, 'host')
        card_host = 'localhost' if v == 0 else v
        self.card = FaceRecognize(card_host,
                                  self.__get_v(STARTUP_SECTION_CARD, 'port', 9998),
                                  self.__get_v(STARTUP_SECTION_CARD, 'output'))
        self.__http_port = self.__get_v(STARTUP_SECTION_RUN, 'http_port', 80)

    @property
    def is_token_server(self):
        return self.__is_token_server

    @property
    def is_create_menu(self):
        return self.__is_create_menu

    @property
    def sandbox(self):
        return self.__sandbox

    @property
    def wx_test(self):
        return self.__wx_test

    @property
    def start_face_server(self):        # 是否启用人脸识别服务器。 1 是； 0 否
        return self.__start_face_server

    @property
    def start_card_server(self):        # 是否启用读卡器服务器。 1 是； 0 否
        return self.__start_card_server

    @property
    def http_port(self):  # web 服务器端口号 默认 80
        return self.__http_port

    @property
    def ws_port(self):  # web socket 服务器端口号 默认 31298
        ws = self.__get_v(STARTUP_SECTION_RUN, 'ws_port', 31298)
        return ws

    @property
    def http_host(self):  # web 服务器 ip， 默认 127.0.0.1
        v_ = self.__get_v(STARTUP_SECTION_RUN, 'http_host')
        return '127.0.0.1' if v_ == 0 else v_

    @property
    def ws_async_mode(self):  # ws 异步模式， 默认 None
        v_ = self.__get_v(STARTUP_SECTION_RUN, 'ws_async_mode')
        return None if v_ == 0 else v_

    @property
    def send_error_email(self):  # 发送异常邮件。 默认 0 否
        v_ = self.__get_v(STARTUP_SECTION_RUN, 'send_error_email')
        return v_

    @property
    def user(self):
        return self.__user

    @property
    def password(self):
        return self.__password

    @property
    def database(self):
        return self.__database

    @property
    def host(self):
        return self.__host

    @property
    def port(self):
        return self.__port

    @property
    def is_mysql(self):
        return self.__is_mysql

    @property
    def redis_pwd(self) -> str:
        return self.__get_v_str(STARTUP_SECTION_REDIS, 'password')

    @property
    def redis_host(self) -> str:
        return self.__get_v_str(STARTUP_SECTION_REDIS, 'host', 'localhost')

    @property
    def redis_port(self) -> int:
        return self.__get_v(STARTUP_SECTION_REDIS, 'port', 6379)

    @property
    def cyt_web_server(self):  # 财艺通 线上服务器
        return self.__get_v(STARTUP_SECTION_RUN, 'cyt_web_server')

    @property
    def task_backup_db(self):       # 启动 备份 数据库 任务
        return self.__get_v(STARTUP_SECTION_TASKS, 'backup_db', 1)

    @property
    def task_wx_token(self):        # 获取 微信公众号 token 任务
        return self.__get_v(STARTUP_SECTION_TASKS, 'wx_token', 1)

    @property
    def gzh_openid(self):  # 微信公众号 openid
        v_ = self.__get_v(STARTUP_SECTION_GZH, 'openid')
        return '' if v_ == 0 else v_

    @property
    def gzh_company_id(self):  # 微信公众号 发送模板消息的 公司id for test ==============
        v_ = self.__get_v(STARTUP_SECTION_GZH, 'company_id')
        return '' if v_ == 0 else v_

    def __get_v(self, s, k, default=0):
        """获取 section=s, key=k的值。 若 s,k 不存在，返回 0. 存在 返回 配置值 v"""
        v = NO_SECTION_OR_KEY
        if self.__startup_cfg:
            v = self.__startup_cfg.get(s, k)

        return v if v != NO_SECTION_OR_KEY else default

    def __get_v_str(self, s, k, default=''):
        """获取 section=s, key=k的值。 若 s,k 不存在，返回 ''. 存在 返回 配置值 v"""
        v = NO_SECTION_OR_KEY
        if self.__startup_cfg:
            v = self.__startup_cfg.get(s, k)

        return str(v) if v != NO_SECTION_OR_KEY else default

    def face_key(self, key):    # 人脸识别 key value 获取
        return self.__get_v(STARTUP_SECTION_FACE, key)

    def face_heat_beat(self):   # 刷脸考勤  是否输出心跳包
        return self.face_key('heart_beat')

    @property
    def show_rsp(self):   # 请求相应
        return self.__get_v(STARTUP_SECTION_RUN, 'show_rsp')

    def show_template(self):   # 显示微信模板内容
        return self.__get_v(STARTUP_SECTION_RUN, 'show_template')

    def get_val(self, val, key=STARTUP_SECTION_RUN):  # 获取配置项，返回值是字符串 或者 数值（能转强转）
        return self.__get_v(key, val)

    def str_val(self, val, key=STARTUP_SECTION_RUN):  # 获取配置项，返回值为字符串
        return self.__get_v_str(key, val)

    def sections(self):   # 查询配置文件中的 段 名称列表
        if not self.__startup_cfg:
            return []
        return self.__startup_cfg.sections()



# -*- coding: utf8 -*-
# =============================================================================
# 文件:  parser_config.py
# 日期:  2021/10/7 14:01
# ini 文件 解析类。
#
# 例如
# ; ini 文件示例， 本行为注释行
# [section_name]
# key = 1  ; 注释内容，英文分号后为注释内容， 整型数字 会转换为 int
# key2 = test   ; 非 整形数字，会返回字符串
#
# 行注释中存在 分号(;) 时， 会忽略注释内容。
# 用法：
#   cfg = JdIniReader(filename)
#   value = cfg.get('section_name', 'key')      # value == 1， 而不是 '1'
#   or
#   val_list = cfg.get('section_name')     # val_list 为 [('key', 1), ('key2', 'test')]
#
# ConfigParser 会将所有的 key 都转换为 小写
# =============================================================================
import os

from configparser import ConfigParser
from threading import Lock
from .tools import JdTools


__version__ = '1.3'


def lock(func):
    def wrapper(self, *args, **kwargs):  # self变量,可以调用对象的相关方法
        with Lock():
            return func(self, *args, **kwargs)

    return wrapper


NO_SECTION_OR_KEY = '<no_section_or_key>'


# 解决ConfigParser模块读取、写入文件时自动转换为小写字符的问题
# 第二种解决方案是自己写一个MyConfigParser，继承自ConfigParser，重写一下optionxform()方法：
class MyConfigParser(ConfigParser):
    def __init__(self, defaults=None):
        ConfigParser.__init__(self, defaults=defaults)

    def optionxform(self, option_str):
        return option_str


class JdIniReader(object):
    """
    读写ini配置文件
    """

    def __init__(self, path):
        self.path = path

        if not os.path.exists(self.path):
            raise IOError('file {} not found!'.format(self.path))
        try:
            self.cf = MyConfigParser()
            self.cf.read(self.path, encoding='UTF-8')
        except Exception as e:
            raise IOError(e)

    def get(self, section, key):
        """读取配置文件数据"""

        if self.cf.has_section(section) and key in self.cf.options(section):
            val = self.cf.get(section, key)
            return JdIniReader.__trim_del_comment(val)
        else:
            return NO_SECTION_OR_KEY

    def sections(self):
        """返回可用的section的列表; 默认section不包括在列表中"""
        return self.cf.sections()

    def get_section(self, section):
        """读取配置文件中某个section下的所有key和value， 返回值 [(k:v),(k:v),...]"""
        lst = self.cf.items(section)
        ret = list()
        for i in lst:
            ret.append((i[0], JdIniReader.__trim_del_comment(i[1])))
        return ret

    @lock
    def set(self, section, key, value):
        """向配置文件写入数据"""
        self.cf.set(section, key, value)
        with open(self.path, 'w', encoding='utf-8') as f:
            self.cf.write(f)

    @staticmethod
    def __trim_del_comment(val):
        # 过滤空格， 删除 ; 或 # 号及后面的所有注释字符
        i = val.find(';')
        j = val.find('#')
        if i >= 0 and j >= 0:
            m = min(i, j)
        elif i >= 0:
            m = i
        elif j >= 0:
            m = j
        else:
            m = -1  # 分号和井号都没有
        if m >= 0:
            val = val[0: m]
        v2 = JdTools.trim(val)
        try:
            v2 = int(v2)    # 若可以转换为 int，则转换，否则原样返回
        except ValueError:
            pass
        return v2



# -*- coding:utf-8 -*-
# 文件:  rsa_pwd.py
# 日期:  2021/10/27 15:45
# 描述:  生成RSA 公钥、私钥， 使用 RSA 解密密码
import base64
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization    # , hashes
import binascii
from .tools import singleton


__version__ = '3.2'

__all__ = [
    'remove_rsa_public_header', 'remove_rsa_private_header',
    'jd_decrypt', 'JdRsa'
]


@singleton
class JdRsa(object):

    def __init__(self):
        self.pub_file = ''
        self.pri_file = ''

    def set_file(self, public, private):
        self.pub_file = public
        self.pri_file = private

    def generate_rsa_key(self):
        # 生成 RSA 公钥 和 私钥

        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend())
        public_key = private_key.public_key()

        # store private key
        private_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )
        with open(self.pri_file, "wb") as f:
            f.write(private_pem)

        # store public key
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        with open(self.pub_file, "wb") as f:
            f.write(public_pem)

    def get_rsa_public_key(self):
        # 从文件获取 RSA 公钥
        with open(self.pub_file) as f:
            pub_key_ = f.read()
            # pub_key_ = remove_rsa_public_header(pub_key_)
        return pub_key_

    def encrypt(self, msg):
        # 加密（encrypt）
        with open(self.pub_file, 'rb') as f:
            key = f.read()
            public_key = serialization.load_pem_public_key(
                key,
                backend=default_backend()
            )

        encrypted = public_key.encrypt(
            msg.encode('utf-8'),
            padding.PKCS1v15()
            # padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA1()), algorithm=hashes.SHA1(), label=None)
        )
        cipher_text = base64.b64encode(encrypted).decode('utf-8')

        return cipher_text

    def decrypt(self, encrypt_text):
        # 解密（decrypt）

        with open(self.pri_file, 'rb') as f:
            key = f.read()
            private_key = serialization.load_pem_private_key(
                key,
                password=None,
                backend=default_backend()
            )

        encrypted = base64.b64decode(encrypt_text)
        original_message = private_key.decrypt(
            encrypted,
            padding.PKCS1v15()
            # padding.OAEP(mgf=padding.MGF1(algorithm=padding.PKCS1v15()), algorithm=hashes.SHA1(), label=None)
        )

        text = original_message.decode('utf-8')
        return text


def remove_rsa_private_header(rsa_private_key):
    if isinstance(rsa_private_key, bytes):
        rsa_private_key = rsa_private_key.decode('utf-8')
    rsa_private_key = rsa_private_key.replace("-----BEGIN RSA PRIVATE KEY-----", '')\
        .replace("-----END RSA PRIVATE KEY-----", '').replace('\n', '')
    return rsa_private_key


def remove_rsa_public_header(rsa_private_key):
    if isinstance(rsa_private_key, bytes):
        rsa_private_key = rsa_private_key.decode('utf-8')
    rsa_private_key = rsa_private_key.replace("-----BEGIN PUBLIC KEY-----", '')\
        .replace("-----END PUBLIC KEY-----", '').replace("\n", '')
    return rsa_private_key


def jd_decrypt(pwd):    # RSA解密，捕获异常
    try:
        jd_rsa = JdRsa()
        password = jd_rsa.decrypt(pwd)
    except (TypeError, ValueError, binascii.Error):
        return False, ''    # 解密错误
    return True, password



# -*- coding: utf8 -*-
import datetime
from .tools import singleton

__version__ = '1.3'
__all__ = ['log', 'cmd', 'JdCmd']


@singleton
class JdCmd(object):
    def __init__(self):
        self._cmd_file = ''  # 输出文件
        self._log_file = ''  # 输出文件
        self._std_out = True    # 输出到屏幕

    @property
    def cmd_file(self):
        return self._cmd_file

    @property
    def log_file(self):
        return self._log_file

    @cmd_file.setter
    def cmd_file(self, new_file):
        self._cmd_file = new_file

    @log_file.setter
    def log_file(self, new_file):
        self._log_file = new_file

    @property
    def output(self):
        return self._std_out

    @output.setter
    def output(self, val):
        self._std_out = val


def log(*args, **kwargs):
    """
    写日志文件，并在 屏幕输出日志
    :param args:        普通参数
    :param kwargs:      key value, 字典
    :return:
    """
    today = datetime.datetime.today()
    date_str = datetime.datetime.strftime(today, '%Y-%m-%d %H:%M:%S')
    ms = '%03d' % (today.microsecond // 1000)
    formatted = f'[{date_str},{ms}]'
    jd_cmd = JdCmd()
    if jd_cmd.output:  # 通过配置文件 控制 是否 输出到屏幕
        print(formatted, *args, **kwargs)
    if jd_cmd.log_file:
        with open(jd_cmd.log_file, 'a', encoding='utf-8') as f:
            print(formatted, *args, file=f, **kwargs)


def cmd(*args, **kwargs):
    """
    写日志文件，并在 屏幕输出日志
    :param args:        普通参数
    :param kwargs:      key value, 字典
    :return:
    """
    today = datetime.datetime.today()
    date_str = datetime.datetime.strftime(today, '%Y-%m-%d %H:%M:%S')
    ms = '%03d' % (today.microsecond // 1000)
    formatted = f'[{date_str},{ms}]'
    jd_cmd = JdCmd()
    print(formatted, *args, **kwargs)
    if jd_cmd.cmd_file:
        with open(jd_cmd.cmd_file, 'a', encoding='utf-8') as f:
            print(formatted, *args, file=f, **kwargs)



# -*- coding:utf-8 -*-
# 文件:  jd_check.py
from .validate import verify
from ..tools import JdTools
from .errcode import *


class JdCheck(object):
    """
    输入参数校验， 用于 web 请求参数校验
    """
    @staticmethod
    def verify(obj, **kwargs):  # 参数校验， 用法说明，请参考 verify 函数
        return verify(obj, **kwargs)

    # =========================================================================
    # 以下校验方式 兼容 旧项目使用， 新项目请统一使用 verify
    # =========================================================================

    @staticmethod
    def decimal(obj, field, description, pt=2):
        """
        校验 obj[field] 是否可以转成 Decimal， 且 小数点 后面的位数 为 pt
        若转换成功 obj[field] 已经是 Decimal 类型
        :param obj:     被检查对象
        :param field:   被检查字段
        :param description:    字段 中文名称
        :param pt:      允许小数点后的位数
        :return:    成功返回 {'code': 0, 'msg': 'ok'}， 失败，code 不为 0
        """
        dd = JdTools.decimal(obj.get(field), pt=pt + 1)
        passed, obj[field] = JdTools.decimal_check(obj.get(field))
        if not passed:
            return {"code": ERR_PUB_NUM, "msg": ERR_PUB_NUM_MSG % description}
        if obj[field] != dd:
            return {"code": ERR_PUB_NUM_PT2, "msg": ERR_PUB_NUM_PT2_MSG % description}

        return {'code': 0, 'msg': 'ok'}

    @staticmethod
    def decimal_ex(obj, **kwargs):
        """
        校验 obj 中多个字段是否可以转为 Decimal 类型
        :param obj:     被校验参数，dict
        :param kwargs:  被校验字段，dict, {字段名称='中文字段名称'}
        :return:  成功 - {'code': 0, 'msg': 'ok'}
                   失败 - {'code': 706, 'msg': 错误信息}
        用法： JdCheck.decimal_ex(obj, real_fee='实收学费', total='应收学费')
        """
        for k, v in kwargs.items():
            vv = obj.get(k)

            dd = JdTools.decimal(vv, pt=3)
            passed, obj[k] = JdTools.decimal_check(vv)
            if not passed:
                return {"code": ERR_PUB_NUM, "msg": ERR_PUB_NUM_MSG % v}
            if obj[k] != dd:
                return {"code": ERR_PUB_NUM_PT2, "msg": ERR_PUB_NUM_PT2_MSG % v}

        return {'code': 0, 'msg': 'ok'}

    @staticmethod
    def decimal_gte0(obj, **kwargs):
        """obj 中字段需为 Decimal，且 大于等于0"""
        for k, v in kwargs.items():
            vv = obj.get(k)

            dd = JdTools.decimal(vv, pt=3)
            passed, obj[k] = JdTools.decimal_check(vv)
            if not passed:
                return {"code": ERR_PUB_NUM, "msg": ERR_PUB_NUM_MSG % v}
            if obj[k] != dd:
                return {"code": ERR_PUB_NUM_PT2, "msg": ERR_PUB_NUM_PT2_MSG % v}

            if dd and dd < 0:
                return {"code": ERR_PUB_GT_EQ_0, "msg": ERR_PUB_GT_EQ_0_MSG % v}

        return {'code': 0, 'msg': 'ok'}

    @staticmethod
    def int_gte0(obj, **kwargs):
        """obj 中字段需为 int，且 大于等于0， 输入值，不能转换为 int时，不报错，默认为0"""
        for k, v in kwargs.items():
            vv = obj.get(k)
            _, obj[k] = JdTools.int(vv)
            if obj[k] < 0:
                return {"code": ERR_PUB_GT_EQ_0, "msg": ERR_PUB_GT_EQ_0_MSG % v}

        return {'code': 0, 'msg': 'ok'}

    @staticmethod
    def int_gte0_err(obj, **kwargs):
        """obj 中字段需为 int，且 【大于等于0】。 输入值，不能转换为 int时，【报错】. 【输入 None 不报错】"""
        for k, v in kwargs.items():
            vv = obj.get(k)
            if vv is None:
                continue
            passed, obj[k] = JdTools.int(vv)
            if not passed:
                return {"code": ERR_PUB_GT_EQ_0, "msg": ERR_PUB_GT_EQ_0_MSG % v}
            if obj[k] < 0:
                return {"code": ERR_PUB_GT_EQ_0, "msg": ERR_PUB_GT_EQ_0_MSG % v}

        return {'code': 0, 'msg': 'ok'}

    @staticmethod
    def str(obj, **kwargs):
        """
        对 obj 中 字符串类型的参数校验长度， 参数可以不填，若填写 不能超长
        :param obj:     被校验参数，dict
        :param kwargs:  被校验字段，dict, {字段名称='长度，中文字段名称'}
        :return:  成功 - {'code': 0, 'msg': 'ok'}
                   失败 - {'code': 700, 'msg': 错误信息}
        用法： JdCheck.str(row, paper_receipt='15,收据号', remark='240,备注')
        """
        for k, v in kwargs.items():
            arr = v.split(',')
            _, length = JdTools.int(arr[0])
            tip = JdTools.trim(arr[1]) or k

            if k not in obj:  # key 未传递
                continue
            obj[k] = JdTools.trim(obj.get(k))
            if obj[k] and len(str(obj[k])) > length:
                return {'code': ERR_PUB_LEN, 'msg': ERR_PUB_LEN_MSG % (tip, length)}

        return {'code': 0, 'msg': 'ok'}

    @staticmethod
    def date(obj, **kwargs):
        """
        校验日期格式是否正确。
        :param obj:         被校验 对象, dict
        :param kwargs:      被校验字段，dict, {字段名称='中文字段名称'}
        :return:    {code: 错误码， msg： 错误提示}
        """
        tip = '“%s”格式错误【%s】，期望 年-月-日！'
        for k, v in kwargs.items():
            dt = obj.get(k)
            try:
                obj[k] = JdTools.str_to_datetime(JdTools.trim(dt), True)
            except (ValueError, TypeError):
                return {'code': ERR_DATE_FMT, 'msg': tip % (v, dt)}
        return {'code': 0, 'msg': 'ok'}



# -*- coding: utf8 -*-
# 文件:  logs.py
# 日期:  2021/11/28 17:42
import logging
from logging.handlers import RotatingFileHandler


__version__ = '1.1.2'
__all__ = ["jd_config_logger", "JdRequestFormatter", "jd_set_logger_file_handler",
           "JD_LOG_FMT", "logger"
           ]


class JdRequestFormatter(logging.Formatter):  # 自定义格式化类
    def format(self, record):
        """每次生成日志时都会调用, 该方法主要用于设置自定义的日志信息
        :param record 日志信息"""
        try:
            from flask import request

            record.url = request.path  # 获取请求的 url url_rule
            record.ip = request.remote_addr  # 获取客户端的ip
        except (RuntimeError, ImportError):
            record.url = '.'
            record.ip = '.'

        return super().format(record)  # 执行父类的默认操作


LOG_FORMAT = ('%(levelname) -10s %(asctime)s  %(name) -50s %(funcName) '
              '-35s %(lineno) -5d: %(message)s')

JD_LOG_FMT = '[%(asctime)s] %(ip)s -- %(url)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'


# flask 中的运行错误也会记录到flask.app的日志中
def jd_config_logger(log, filename=None, count=5, size=50):
    """配置 flask 日志"""
    # 创建flask.app日志器
    jd_logger = log
    # 设置全局级别
    jd_logger.setLevel('DEBUG')

    # 创建控制台处理器
    console_handler = logging.StreamHandler()

    # 给处理器设置输出格式
    file_formatter = JdRequestFormatter(JD_LOG_FMT)

    # fmt = '[%(asctime)s] %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]'
    # console_formatter = logging.Formatter(fmt)
    console_handler.setFormatter(file_formatter)

    # 日志器添加处理器
    jd_logger.addHandler(console_handler)

    # 创建文件处理器
    if filename:
        jd_set_logger_file_handler(log, filename, count, size)


def jd_set_logger_file_handler(log, filename, count=5, size=50):
    # 创建文件处理器
    jd_logger = log
    file_handler = RotatingFileHandler(filename=filename, maxBytes=size * 1024 * 1024, backupCount=count,
                                       encoding='utf-8')

    # 给处理器设置输出格式
    file_formatter = JdRequestFormatter(JD_LOG_FMT)

    # 给处理器设置输出格式
    file_handler.setFormatter(file_formatter)
    # 单独设置文件处理器的日志级别
    file_handler.setLevel('DEBUG')

    # 日志器添加处理器
    jd_logger.addHandler(file_handler)


logger = logging.getLogger('app')
jd_config_logger(logger)



# -*- coding: utf8 -*-
# =============================================================================
# 文件:  validate.py
# 日期:  2021/4/29 11:07
# 修改记录：
#   2022-05-12  WXG  verify date格式 支持 参数 no-hms, 时分秒为 00:00:00
#   2022-06-05  WXG  verify in参数增加缺少结束符 ']' 的校验。版本 1.1.3
#   2022-07-08  WXG  verify 增加ppt，自定义错误提示。版本 1.1.4
# =============================================================================
from collections import defaultdict
from abc import ABCMeta, abstractmethod

from ..tools import JdTools, JdDate
from .errcode import *


__version__ = "1.1.7"
__author__ = "WXG"
__all__ = ['verify', 'validation']


class Validator(object):
    """
    参数校验类， 纯虚基类
    """

    __metaclass__ = ABCMeta
    err_message = ""

    @abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class Range(Validator):
    """
    范围： [start, end], inclusive 是否包含，为假时的取值范围：(start, end)
    # Example:
        validations = {
            "field": [Range(0, 10)]
        }
        passes = {"field": 10}
        fails = {"field" : 11}
    """

    def __init__(self, start, end, inclusive=True):
        self.start = start
        self.end = end
        self.inclusive = inclusive
        self.err_message = f"取值范围必须在 [{start}-{end}]"

    def __call__(self, value):
        if self.inclusive:
            return self.start <= value <= self.end
        else:
            return self.start < value < self.end


class Length(Validator):
    """
    长度校验： 最小长度、最大长度 二者必填其一
    # Example:
        validations = {
            "field": [Length(0, maximum=5)]
        }
        passes = {"field": "hello"}
        fails  = {"field": "hello world"}
    """

    err_messages = {
        "maximum": "最大长度为 {0}",
        "minimum": "最小长度为 {0}",
        "range": "长度必须在 {0}-{1} 之间",
        "equal": "长度必须为 {0}"
    }

    def __init__(self, minimum, maximum=0):
        if not minimum and not maximum:
            raise ValueError("Length must have a non-zero minimum or maximum parameter.")
        if minimum < 0 or maximum < 0:
            raise ValueError("Length cannot have negative parameters.")

        self.minimum = minimum
        self.maximum = maximum
        if minimum and maximum:
            if minimum == maximum:
                self.err_message = self.err_messages["equal"].format(minimum)
            else:
                self.err_message = self.err_messages["range"].format(minimum, maximum)
        elif minimum:
            self.err_message = self.err_messages["minimum"].format(minimum)
        elif maximum:
            self.err_message = self.err_messages["maximum"].format(maximum)

    def __call__(self, value):
        length = len(str(value))
        if self.maximum:
            return self.minimum <= length <= self.maximum
        else:
            return self.minimum <= length


class GreaterThan(Validator):
    """
    大于(等于):
    # Example:
        validations = {
            "field": [GreaterThan(10)]
        }
        passes = {"field": 11}
        fails = {"field" : 10}
    """

    def __init__(self, lower_bound, inclusive=False):
        self.lower_bound = lower_bound
        self.inclusive = inclusive
        self.err_message = "必须大于%s%s" % ('等于' if inclusive else '', lower_bound)

    def __call__(self, value):
        if self.inclusive:
            return value >= self.lower_bound
        else:
            return value > self.lower_bound


class LessThan(Validator):
    """
    小于（等于）
    # Example:
        validations = {
            "field": [LessThan(10)]
        }
        passes = {"field": 9}
        fails = {"field" : 10}
    """

    def __init__(self, upper_bound, inclusive=False):
        self.upper_bound = upper_bound
        self.inclusive = inclusive
        self.err_message = "必须小于%s%s" % ('等于' if inclusive else '', upper_bound)

    def __call__(self, value):
        if self.inclusive:
            return value <= self.upper_bound
        else:
            return value < self.upper_bound


class In(Validator):
    """
    取值范围， 判断是否属于某 集合 （列表）
    # Example:
        validations = {
            "field": [In([1, 2, 3])]
        }
        passes = {"field":1}
        fails  = {"field":4}
    """

    def __init__(self, collection):
        self.collection = collection
        self.err_message = "取值范围不在 %r 中" % collection

    def __call__(self, value):
        return value in self.collection


class OpValidator(Validator):
    def __call__(self, value):
        return True  # 校验不返回失败


class Optional(OpValidator):
    """可选， 可选参数"""
    pass


class Required(OpValidator):
    """必填"""
    pass


class Must(OpValidator):
    """必填， 日期时， None 和 ''空字符 报错 """
    pass


class NoHms(OpValidator):
    """无 时分秒， 日期时, 只有 年-月-日， 时分秒 赋值为 00:00:00 """
    pass


class Default(OpValidator):
    """默认值"""
    def __init__(self, default):
        self.default = default


class Point(OpValidator):
    """小数点位数"""
    def __init__(self, point):
        self.point = point


class Any(OpValidator):
    """任意取值， 同类型内，比如 int / decimal"""
    pass


class Prompt(OpValidator):
    """提示语，代替默认提示"""
    def __init__(self, ppt):
        self.ppt = ppt


class _WRONG(object):
    pass


def _set_default(default, who):
    """
    计算默认值， ds
        ds 默认值， str 默认值为 空字符串("")， date/time -- None,
        decimal/int -- 0, dateRange -- 不支持
    :param default:     默认值， s for ds， 0/1/5.6 for int, decimal: d0, d5.6
    :param who:         参数类型：int, date, str, ...
    :return:    默认值 or _WRONG
    """
    dd = default
    if default in ["s", "0"]:  # 默认值 ds
        if who in ['int', 'i', 'D', 'decimal']:
            dd = 0
        elif who in ['s', 'str']:
            dd = ""
        elif who in ['t', 'dateRange']:
            dd = _WRONG
        elif who in ['d', 'date', 't', 'time']:     # time 也支持 ds
            dd = None
    return dd


def validation(s, key, who):
    """
    字符串 转 参数校验对象
    """
    validations = []

    arr = JdTools.split(s, ',')
    name = arr[0]
    a2 = arr[1:]

    mode = 0    # 0-无任何模式， 1-in 模式，当前解析 in 语句
    in_list = []
    errors = defaultdict(list)
    for i in a2:
        if mode == 1:  # 1-in 模式，当前解析 in 语句
            if ']' in i:
                mode = 0  # 退出 in 模式
                m = JdTools.split(i, ']')
                in_list.append(m[0])

                in_arr = []
                for m in in_list:
                    if who in ['i', 'int']:
                        passed, v = JdTools.int(m)
                        if not passed:
                            errors[key].append(f"取值范围(In) 转换为int失败[ {m[0]} ]！")
                        in_arr.append(v)
                    else:
                        in_arr.append(m)
                validations.append(In(in_arr))
            else:
                in_list.append(i)
        elif i == 'opt':  # 可选参数
            validations.insert(0, Optional)
        elif i == 'must':  # 必填参数，默认为 必填
            validations.append(Must)
        elif i == 'no-hms':     # 时分秒为 00:00:00
            validations.append(NoHms)
        elif i[0] == 'd':  # default，默认值
            val = i[1:]
            success = True
            if val in ['s', '0']:  # ds - 默认值
                default = _set_default(val, who)
                if default == _WRONG:
                    errors[key].append("time/dateRange不支持默认值！")
                    success = False
            elif '.' in val:
                success, default = JdTools.decimal(val)
                if not success:
                    errors[key].append("默认值不合法！")
            else:
                success, default = JdTools.int(val)
                if not success:
                    errors[key].append("默认值不合法，应该为整数！")
            if success:
                validations.append(Default(default))
        elif i[:3] == 'ppt':    # 提示语
            arr_ppt = JdTools.split(i, 'ppt')
            if not arr_ppt[1]:
                errors[key].append("提示语不能为空！")
            else:
                validations.append(Prompt(arr_ppt[1]))
        elif i[0] == 'p':  # 小数点后几位
            success, val = JdTools.int(i[1])
            if not success:
                errors[key].append("p后应为整数！")
            else:
                validations.append(Point(val))
        elif i[0] == '<':   # 小于
            if i[1] == '=':  # 小于等于 <=
                success, operand = JdTools.int(i[2:])
                if not success:
                    errors[key].append("取值范围书写错误！")
                else:
                    validations.append(LessThan(operand, inclusive=True))
            else:  # 小于 <
                success, operand = JdTools.int(i[1:])
                if not success:
                    errors[key].append("取值范围书写错误！")
                else:
                    validations.append(LessThan(operand))
        elif i[:2].lower() == 'in':  # in 语句， 例如 state=状态, in[0,1]
            mode = 1
            m = JdTools.split(i, '[')
            in_list = [m[-1]]
        else:  # 范围
            if '-' in i:
                start, end = JdTools.split(i, '-')
            else:
                start, end = 0, i
            success1, start = JdTools.int(start)
            success2, end = JdTools.int(end)
            if not success1:
                errors[key].append("取值范围最小值书写错误！")
            if not success2:
                errors[key].append("取值范围最大值书写错误！")
            if success1 and success2:
                validations.append(Length(start, end))
                if who in ['s', 'str'] and start == 0:  # 可选字符
                    validations.insert(0, Optional)

    if mode == 1:
        errors[key].append(f"in 参数缺少结束符']'！")

    # 第一次分隔 取值范围校验
    if who in ['int', 'i', 'D', 'decimal']:
        first = name
        arr = JdTools.split(first, '>')
        if len(arr) > 1:  # 存在大于号（'>'）
            if arr[1][0] == '=':  # op = 'gte' 大于等于 >=
                success, operand = JdTools.int(arr[1][1:])
                if not success:
                    errors[key].append("取值范围书写错误！")
                else:
                    validations.append(GreaterThan(operand, inclusive=True))
            else:  # op = 'gt'  # 大于  >
                success, operand = JdTools.int(arr[1])
                if not success:
                    errors[key].append("取值范围书写错误！")
                else:
                    validations.append(GreaterThan(operand))
        else:
            arr = JdTools.split(first, 'any')
            if len(arr) > 1:  # op = 'any'  # 能转换为 int/decimal 的任何值
                validations.append(Any)
            else:
                validations.append(GreaterThan(0, inclusive=True))  # int, decimal 默认 >= 0
    name = arr[0]

    return name, validations, errors  # name-key对象的中文名称, errors-错误列表


def verify(obj, **kwargs):
    """
    参数校验
    :param obj:
    :param kwargs:
        key     type                opt    length  scope  default point  in   ppt
        name    s for str           √     √      -      √      -      √
                d for date          √     -       -      √      -      -
                D for decimal       √     -       √     √      √     √
                i for int           √     -       √     √      -      √   √
                t for time          √     -       -      √      -      -
                dateRange           √     -       -      -       -      -

        用法示例：
        ret = JdCheck.verify(obj, s=dict(name='客户名称, 1-50', short_name='简名, 1-10', person='联系人, 20',
            remark='备注, 140'))

        ***********************************************************************
        规则：
        优先级  opt > must > d(default)
        1. 【opt】 为 可选参数， 没有对应 key 时， 写库不传递该 key ； 有对应
           key 时，opt 不起作用。
           - 优先级 opt < default(d)  即有默认值时，opt 无效
           - 不加 opt，按必填参数处理 ★★★。
        2. 【must】 为 必填参数。 没有对应 key 时 报错。 日期时， None 值报错
        3. 【d】 为 默认值(default)， 未传递 key 或者 值为 '', None 时，赋默认值。
            - ds 默认值， str 默认值为 空字符串("")， date -- None,
              decimal/int -- 0, time / dateRange -- 不支持
        4. scope 取值范围包括： >=0(int, decimal 默认范围), >0, >x, >=x;
           any (任意取值)。
        5. 【p】point, 默认为 p2 小数点后2位，只对 decimal 有效 。
        6. 【length】 目前对 str 类型有效， 长度区间 [min, max]， min默认为0 。
        7. ppt:  错误提示语， 不加 ppt 按 系统原有提示。
        ***********************************************************************
        【str】
        name='客户名称, 1-50'     必填，长度 1-50。 长度不在范围内，报错
        person='联系人, 20        可选，长度最大 20。 长度超过最大值，报错
        person='联系人, 20, opt   允许没有 person 参数，若有 person，按照 上述方式校验。

        【decimal】  不能转为 Decimal 报错。None 或者 '' 转为 0。
        fee='学费'                大于等于0
        fee='学费 >= 0'           大于等于0
        fee='学费>0'              必填，大于0， 默认小数点后2位
        fee='学费 >= 0, p2'       大于等于0， 小数点后2位
        fee='学费 any'            可以转换为 Decimal 的任何值

        【date】  不能转为 日期 报错。无 must 时，None 或者 '' 转为 None。 有 must 时， None/'' 报错
        date='充值日期'           必填。 日期 按 年-月-日 时:分:秒 赋值，无时分秒，填写当前时分秒
        date='充值日期, opt'      可选。
        date='充值日期, must'     必填，None/'' 值报错。
        date='考勤日期, no-hms'   必填， 时分秒 为 0:0:0

        【time】
        time='演出时长'          必填，    时间字符串格式 HH:MM:SS，例如：'00:05:34' 为 5分34秒
        time='演出时长，opt'     可选

        【int】  不能转为 int 报错。None 或者 '' 报错。
        times='课时'              大于等于0
        times='课时 >0'           大于0
        times='课时 any'          可以转为 int 的任何值
        times='课时,d2'           默认值为2， 若填写，按之前规则校验（这里为大于等于0）
        times='课时,opt'          可选，没有该参数，不报错。有该参数，按照 上述规则校验
        state='状态, in[0,1]'     取值范围属于列表

        【dateRange】 可转换为日期范围， 否则报错
        dataVal='2022-1-1 - 2022-1-31'      obj中生成 d1, d2，且 d1=2022-1-1(datetime类型) d2=2022-1-31
                                  date_from_to='从 2022-01-01 至 2022-01-31'
    """
    extra_tip = ''
    for k, v in kwargs.items():
        if k == 'tip':
            extra_tip = v or ''  # 附加提示信息，放在开头。 比如 对续班子表校验时，可输入 学员姓名
        else:
            if not isinstance(v, dict):  # 异常处理
                return {'code': ERR_SYNTAX, 'msg': ERR_SYNTAX_MSG % f"{k}={v}"}
            for k2, v2 in v.items():    # 遍历 date， int 等 具体类型的校验
                ret = _validate_single_key(k2, v2, obj, k, extra_tip)
                if ret['code'] != 0:
                    return ret

    return {'code': 0, 'msg': 'ok'}


def _validate_single_key(key, s, dictionary, who, extra_tip):
    """
     将 key 对应 规则字符串 s，转为 Validator 列表。
     被校验者：  dictionary
     类型： who , 字符串类型，取值  s, str, i, int, date, decimal 等
     额外提示： extra_tip
    """
    tip, validations, errors = validation(s, key, who)
    name = f'{extra_tip}“{tip}”' if extra_tip else f'{tip}'
    if errors:
        return {'code': ERR_VERIFY_ERROR, 'msg': f'{extra_tip}“{tip}”，{errors[key][0]}'}

    # 读取特殊 校验 对象
    special = {'ppt': {'has': False, 'val': ''},        # ppt， 提示
               'default': {'has': False, 'val': ''},    # 默认值
               'point': {'has': False, 'val': ''},      # 小数点后位数
               'add_hms': {'has': True},                # 是否添加 时分秒 for date
               }
    for validator in validations:
        if isinstance(validator, Prompt):  # 自定义提示
            special['ppt'].update({'has': True, 'val': validator.ppt})
        elif isinstance(validator, Default):  # 默认值
            special['default'].update({'has': True, 'val': validator.default})
        elif isinstance(validator, Point):  # 小数点
            special['point'].update({'has': True, 'val': validator.point})
        elif validator == NoHms:  # 不需要添加 年月日
            special['add_hms'].update({'has': False})

    # 默认值
    if key not in dictionary or dictionary[key] == '' or dictionary[key] is None:
        if special['default']['has']:
            dictionary[key] = special['default']['val']

    if Optional in validations and key not in dictionary:  # 可选参数
        return {'code': 0, 'msg': 'ok'}
    if Required in validations and key not in dictionary:  # 必填参数
        return {'code': ERR_VERIFY_NEED_PARAM, 'msg': f'请输入【{name}】'}
    if Must in validations:
        if key not in dictionary:  # 必填参数
            return {'code': ERR_VERIFY_NEED_PARAM, 'msg': f'请输入【{name}】'}
        else:
            if who in ['date', 'd']:
                if dictionary[key] is None or dictionary[key] == '':
                    return {'code': ERR_VERIFY_NEED_PARAM, 'msg': f'请输入【{name}】'}

    if key not in dictionary:
        return {'code': ERR_PUB_NEED, 'msg': special['ppt']['val'] or ERR_PUB_NEED_MSG % name}

    val = dictionary[key]
    if who in ['s', 'str']:
        dictionary[key] = JdTools.trim(str(val or ''))
    elif who in ['D', 'decimal']:
        pt = 2  # for decimal，小数点后的最大位数
        if special['point']['has']:
            pt = special['point']['val']

        passed, dictionary[key] = JdTools.decimal_check(val, pt)
        if not passed:
            return {"code": ERR_PUB_NUM, "msg": special['ppt']['val'] or ERR_PUB_NUM_MSG % name}
        dec = JdTools.decimal(val, pt=pt + 1)
        if dictionary[key] != dec:
            return {"code": ERR_PUB_NUM_PT2, "msg": f'{name} 最多 {pt} 位小数！'}
        # 校验 decimal 数值 不能超过 7 位数字
        if dec > 9999999:
            return {'code': ERR_VERIFY_VAL_EXCEED, 'msg': special['ppt']['val'] or f'{name} 数值越限，最多7位（除小数外）'}
    elif who in ['d', 'date']:
        add_hms = special['add_hms']['has']
        try:
            dictionary[key] = JdTools.str_to_datetime(val, add_hms=add_hms)
        except (ValueError, TypeError):
            return {'code': ERR_DATE_FMT, 'msg': special['ppt']['val'] or f'{name} 格式错误【{val}】，期望 年-月-日！'}
    elif who in ['t', 'time']:      # 时间 时分秒
        try:
            dictionary[key] = JdTools.str_to_time(val)
        except (ValueError, TypeError):
            return {'code': ERR_TIME_FMT, 'msg': special['ppt']['val'] or f'{name} 格式错误【{val}】，期望 HH:MM:SS！'}
    elif who in ['i', 'int']:
        passed, dictionary[key] = JdTools.int(val)
        if not passed:
            return {"code": ERR_PUB_NUM_INT, "msg": special['ppt']['val'] or ERR_PUB_NUM_INT_MSG % name}
    elif who in ['dateRange']:
        ret = JdDate.start_end_time('time_range', val)
        if ret['code'] != 0:
            return ret
        dictionary['d1'] = ret['d1']
        dictionary['d2'] = ret['d2']
        dictionary['date_from_to'] = ret['date_from_to']
    else:
        return {"code": ERR_VERIFY_WHO, "msg": f'未知的校验类型 {who}'}

    for validator in validations:
        if validator in [Any, Optional, Required, Must, NoHms]:
            continue
        passed = validator(dictionary[key])
        if not passed:
            return {'code': ERR_VERIFY_ERROR, 'msg': special['ppt']['val'] or f'[{name}] {validator.err_message}'}

    return {'code': 0, 'msg': 'ok'}
